/* $Id: Amazons.java,v 1.10 2006/08/30 08:52:09 orts_mburo Exp $
   
   (c) Michael Buro, licensed under the GPL2
*/

package ggs.ggsa._amazons;

import java.util.*;
import java.awt.*;
import ggs.ggsa.boardgamesvc.*;
import ggs.ggsa.main.*;
import ggs.shared.*;


public class Amazons extends BoardGameServiceClient
{
  public Amazons() { super(new AmazonsGame()); }
  
  public void setDefaultOptions() 
  {
    new AmazonsGame().setDefaultOptions();
  }

}

class AmazonsGame extends MailBoxBoardGame
{
  public String get_name()    { return "Amazons"; }  
  public String get_base_name()    { return "ams"; }
  public BoardGame new_BoardGame() { return new AmazonsGame(); }

  public void setDefaultOptions()
  {
    String sn = get_base_name() + ".";

    super.set_default_options();
    
    Global.options.put2i(sn+"ipw", BGS+"ipw"); // info panel
    Global.options.put2i(sn+"tph", BGS+"tph"); // game window vertical split

    Global.options.put2i(sn+"ge1", BGS+"ge1"); // 3 board window geoms
    Global.options.put2i(sn+"ge2", BGS+"ge2");
    Global.options.put2i(sn+"ge3", BGS+"ge3"); // also synchro geometry

    // playfield and grid color
    Global.options.put2i(sn+"gc",  BGS+"gc");  // grid color

    Global.options.put(sn+"mo",  new Boolean(true));         // show moves (boolean)
    Global.options.put(sn+"la",  new Boolean(true));         // show last move (boolean)

    // empty square color
    Global.options.put(sn+"es0", new Integer(30*65536  + 144*256 + 255));
    Global.options.put(sn+"es1", new Integer(0*65536   + 191*256 + 255));
    
    // hit square color
    Global.options.put(sn+"hs",  new Integer(255*65536 + 160*256 + 122));

    // anti empty square color (not used)
    Global.options.put(sn+"aes0", new Integer(255*65536 + 114 * 256 + 86));
    Global.options.put(sn+"aes1", new Integer(255*65536 + 114 * 256 + 86));

    Global.options.put(sn+"cm", "");
  }

  public static final int HIT = 4;
  
  public static final int GFX_BLACK_DISC   = 4;
  public static final int GFX_WHITE_DISC   = 8;
  public static final int GFX_HIT          = 16;  
  public static final int GFX_BLACK_MOVE   = 32;
  public static final int GFX_WHITE_MOVE   = 64;
  public static final int GFX_BLACK_GHOST1 = 128;
  public static final int GFX_WHITE_GHOST1 = 256;
  public static final int GFX_BLACK_GHOST2 = 512;
  public static final int GFX_WHITE_GHOST2 = 1024;

  public static final int GFX_DISC_MASK = GFX_BLACK_DISC + GFX_WHITE_DISC;

  protected MailBoxGameState new_MailBoxGameState()
  {
    return new AmazonsGameState();
  }

  public boolean has_synchro()     { return true;  }
  public boolean has_komi()        { return true;  }
  public boolean has_rand()        { return true;  }
  public boolean has_rand_type()   { return false; }
  public boolean has_anti()        { return false; }  
  public boolean has_pref_color()  { return true;  }    
  public boolean has_pass()        { return false; } 

  public boolean never_turn()        { return false; }  
  public boolean turned_if_black()   { return false; }
  protected boolean y_flip()         { return true; }
  protected boolean pieces_on_grid() { return false; }
  protected boolean only_squares()   { return true;  }

  protected int border_width()     { return 1; }

  public int board_type_num()      { return 6; }
  public int default_board_type()  { return 10; }

  public int board_type(int index)
  {
    switch (index) {
    case 0:  return 4;
    case 1:  return 6;
    case 2:  return 8;
    case 3:  return 10;
    case 4:  return 12;      
    case 5:  return 14;
    default: return -1;
    }
  }

  public int board_type_index(int type) {
    switch (type) {
    case 4:  return 0;
    case 6:  return 1;
    case 8:  return 2;
    case 10: return 3;
    case 12: return 4;      
    case 14: return 5;
    default: return -1;
    }
  }
  
  public String board_type_string(int index) {
    switch (index) {
    case 0:  return "4x4";
    case 1:  return "6x6";
    case 2:  return "8x8";
    case 3:  return "10x10";
    case 4:  return "12x12";      
    case 5:  return "14x14";
    default: return "?";
    }
  }

  public int width(int type) { return type; }
  
  public int height(int type) { return width(type); }
  
  public byte char2cont(char c)
  {
    switch (c) {
    case '*': return BLACK;
    case 'O': return WHITE;
    case '-': return EMPTY;
    case '+': return HIT;      
    default:
      throw new Error("illegal char in Amazons.char2cont");
    }
  }

  public char cont2char(byte co)
  {
    switch (co) {
    case BLACK: return '*';
    case WHITE: return 'O';
    case EMPTY: return '-';
    case HIT:   return '+';      
      
    default:
      throw new Error("illegal cont in Amazons.cont2char");
    }
  }


  private boolean on_same_line(int ind1, int ind2)
  {
    int x1, y1, x2, y2;
    
    x1 = ind2x(ind1);
    y1 = ind2y(ind1);
    x2 = ind2x(ind2);
    y2 = ind2y(ind2);
    
    return x1 == x2 || y1 == y2 || Math.abs(x1-x2) == Math.abs(y1-y2);
  }


  private static int sign(int d)
  {
    if (d > 0) return 1;
    if (d < 0) return -1;
    return 0;
  }

  private int cont_num(AmazonsGameState state, int cont)
  {
    int n = 0;

    for (int i=0; i < BOARD_SIZE; ++i)
      if (state.sq[i] == cont) ++n;

    return n;
  }
                       
  private boolean empty_line(AmazonsGameState st, int ind1, int ind2)
  {
    if (ind1 == ind2) return false;
    if (!on_same_line(ind1, ind2)) return false;
    int x1, y1, x2, y2;

    x1 = ind2x(ind1);
    y1 = ind2y(ind1);
    x2 = ind2x(ind2);
    y2 = ind2y(ind2);

    int dx = sign(x2-x1);
    int dy = sign(y2-y1);
    int d = dx + dy * DX;
    int p = ind1 + d;

    while (p != ind2) {
      if (st.sq[p] != EMPTY) return false;
      p += d;
    }
  
    return true;
  }

  private boolean no_moves(AmazonsGameState st, int color) {

    for (int i=0; i < BOARD_SIZE; ++i) {
      if (st.sq[i] == color) {
	for (int j=0; j < 8; j++) {
	  if (st.sq[i+D[j]] == EMPTY) return false;
	}
      }
    }

    return true;
  }

  public MailBoxGameState make_move2(String mv)
  {
    // parse move

    Vector<MovePart> v = new Vector<MovePart>();
    String err = parse_move(mv, v);

    if (err != null) return null;
    if (v.size() != 3) return null;

    MovePart mp = (MovePart) v.elementAt(0);
    if (!mp.rest.equals("")) return null;
    int from = mp.square;

    mp = (MovePart) v.elementAt(1);
    if (!mp.rest.equals("")) return null;
    int to = mp.square;

    mp = (MovePart) v.elementAt(2);
    if (!mp.rest.equals("")) return null;
    int arrow = mp.square;

    // check move

    AmazonsGameState state = (AmazonsGameState) current_state();
    
    if (state.sq[from]!= state.to_move) return null;
    if (state.sq[to]  != EMPTY)   return null;
    if (from != arrow && state.sq[arrow] != EMPTY) return null;    
    
    if (!on_same_line(from, to))  return null;
    if (!on_same_line(to, arrow)) return null;  
    if (!empty_line(state, from, to)) return null;

    AmazonsGameState new_state = (AmazonsGameState) state.createCopy();
    
    new_state.sq[to]   = (byte)new_state.to_move;
    new_state.sq[from] = EMPTY;
    if (!empty_line(new_state, to, arrow)) return null;

    new_state.sq[arrow] = HIT;

    int opp = PlayerColor.opponent(new_state.to_move);
    
    if (no_moves(new_state, opp))
      new_state.white_passes += new_state.to_move == BLACK ? 1 : -1;    
    else
      new_state.to_move = opp;
    
    return new_state;
  }

  public void input(String s) { }

  protected void process_mouse_events() {

    boolean ok = false;
    int from = PASS;
    int to = PASS;
    int arrow = PASS;

    synchronized (mev) {
      
      int n = mev.size();
      int i;

//       Global.dbgmsg("PROCESS_MOUSE_EVENTS BEFORE:");
//       for (i=0; i < mev.size(); ++i) {
// 	MouseEvent m = (MouseEvent)mev.elementAt(i);
// 	Global.dbgmsg(m.type + " " + m.square);
//       }
      
      for (i=0; i < mev.size(); ++i) {
	MouseEvent m = (MouseEvent)mev.elementAt(i);
	if (m.type == PRESSED) break;
	mev.removeElementAt(i);
	--i;
	continue;
      }

      ++i;
      
      for (;i < mev.size(); ++i) {
	MouseEvent m = (MouseEvent)mev.elementAt(i);
	if (m.type == RELEASED) break;
	mev.removeElementAt(i);
	--i;
	continue;
      }

      ++i;
      
      for (;i < mev.size(); ++i) {
	MouseEvent m = (MouseEvent)mev.elementAt(i);
	if (m.type == CLICKED) break;
	mev.removeElementAt(i);
	--i;
	continue;
      }

//       Global.dbgmsg("PROCESS_MOUSE_EVENTS AFTER:");
      
//       for (i=0; i < mev.size(); ++i) {
// 	MouseEvent m = (MouseEvent)mev.elementAt(i);
// 	Global.dbgmsg(m.type + " " + m.square);
//       }
      
      if (mev.size() == 0) return;

      AmazonsGameState state = (AmazonsGameState)current_state();

      if (mev.size() >= 1) {
	from = ((MouseEvent)mev.elementAt(0)).square;

	if (state.sq[from] != state.to_move) { reset_mev(); return; }
      }	

      if (mev.size() >= 2) {
	to = ((MouseEvent)mev.elementAt(1)).square;
	if (state.sq[to] != EMPTY) { reset_mev(); return; }
      }	

      if (mev.size() >= 3) {
	arrow = ((MouseEvent)mev.elementAt(2)).square;
	ok = true;
      }
    }

    if (ok) {
      handler.handle_event(new EventMsg(GameWindow.USER_MOVE_ID,
					ind2string(from) + "-" +
					ind2string(to)   + "-" +
					ind2string(arrow)));
      reset_mev();
    }
  }

  protected boolean entire_gfx_cont(MailBoxGameState st, String last_move) {

    AmazonsGameState state = (AmazonsGameState)st;

    for (int i=0; i < BOARD_SIZE; i++) {

      if (state.sq[i] == BORDER) {

        gfx_sq[i] = GFX_BORDER;

      } else if (state.sq[i] == HIT) {

        gfx_sq[i] = GFX_HIT;

      } else {

        if (state.sq[i] == BLACK) 
          gfx_sq[i] = GFX_BLACK_DISC;
        
        else if (state.sq[i] == WHITE) 
          gfx_sq[i] = GFX_WHITE_DISC;

        else
          gfx_sq[i] = GFX_EMPTY;

      }
    }

    if (show_last_move) {
      Vector<MovePart> v = new Vector<MovePart>();
      if (parse_move(last_move, v) == null) {
	gfx_sq[((MovePart)v.elementAt(1)).square] |= GFX_LAST_MOVE;
	gfx_sq[((MovePart)v.elementAt(2)).square] |= GFX_LAST_MOVE;
      }
    }

    int move_flag = ((state.to_move == PlayerColor.BLACK) ?
		     GFX_BLACK_MOVE :
		     GFX_WHITE_MOVE);
    int opp_move_flag = ((state.to_move == PlayerColor.BLACK) ?
			 GFX_WHITE_MOVE :
			 GFX_BLACK_MOVE);
    int ghost1_flag = ((state.to_move == PlayerColor.BLACK) ?
		       GFX_BLACK_GHOST1 :
		       GFX_WHITE_GHOST1);
    int ghost2_flag = ((state.to_move == PlayerColor.BLACK) ?
		       GFX_BLACK_GHOST2 :
		       GFX_WHITE_GHOST2);

    synchronized (mev) {
    
      if (mev.size() == 0) {

	if (show_moves) {
      
	  // no click: mark amazons to move
        
	  for (int i=0; i < BOARD_SIZE; i++) {
          
	    if (state.sq[i] == state.to_move) {
            
	      // move possible?
            
	      for (int j=0; j < 8; j++) 
		if (state.sq[i + D[j]] == EMPTY) {
		  gfx_sq[i] |= opp_move_flag;
		  break;
		}
	    }
	  }
	}
      
      } else if (mev.size() == 1) {

	int coor0 = ((MouseEvent)mev.elementAt(0)).square;

	if (show_moves) {
        
	  // one click: mark amazon moves
	
	  for (int i=0; i < 8; i++) {
          
	    int pos = coor0 + D[i];
          
	    while (state.sq[pos] == EMPTY) {
	      gfx_sq[pos] |= move_flag;
	      pos += D[i];
	    }
	  }
	}
      
	// show ghost1
      
	gfx_sq[coor0] = ghost1_flag;
      
      } else if (mev.size() == 2) {
    
	int coor0 = ((MouseEvent)mev.elementAt(0)).square;
	int coor1 = ((MouseEvent)mev.elementAt(1)).square;

	if (show_moves) {
        
	  // two clicks: mark arrow locations
      
	  for (int i=0; i < 8; i++) {
          
	    int d = D[i];
	    int pos = coor1 + d;
          
	    while (state.sq[pos] == EMPTY || pos == coor0) {
	      gfx_sq[pos] |= move_flag;
	      pos += d;
	    }
	  }
	}
      
	// no amazon on from square, ghost2 on to-square
      
	gfx_sq[coor0] &= (~ghost1_flag) & (~GFX_DISC_MASK);
	gfx_sq[coor1] |= ghost2_flag;
      }
    }
    return true;
  }
  
  protected int last_move_index(String move) {
    Vector<MovePart> v = new Vector<MovePart>();
    if (parse_move(move, v) != null) return -1;
    return (v.elementAt(1)).square;
  }

  public int gfx_cont(MailBoxGameState st, int i) { return GFX_EMPTY; }
  
  protected void paint_gfx_cont(Graphics g, int x, int y, int parity, int gfx_cont) {

    if ((gfx_cont & GFX_BORDER) != 0) return;

    if ((gfx_cont & GFX_HIT) != 0) {

      g.setColor(new Color(Global.options.getInteger(get_base_name()+".hs").intValue()));
      
    } else {

      if (parity == 0) g.setColor(col_empty0); else g.setColor(col_empty1);

    }
    
    g.fillRect(x+1,y+1,gfx_sq_w-1, gfx_sq_h-1);
    int disc_diam = (int)Math.round(gfx_sq_w * 0.85);
    disc_diam &= ~1;
    
    if ((gfx_cont & GFX_BLACK_GHOST1) + (gfx_cont & GFX_WHITE_GHOST1) > 0) {

      Color col;
      int xl = x + (gfx_sq_w - disc_diam)/2;
      int yt = y + (gfx_sq_w - disc_diam)/2;
      
      if ((gfx_cont & GFX_BLACK_GHOST1) != 0) col = col_black; else col = col_white;

      g.setColor(col);
      // g.drawOval(xl, yt, disc_diam, disc_diam);
      Bresenham.drawCircle(g, xl, yt, disc_diam);

    } else if ((gfx_cont & GFX_BLACK_GHOST2) + (gfx_cont & GFX_WHITE_GHOST2) > 0) {

      Color col;
      int xl = x + (gfx_sq_w - disc_diam)/2;
      int yt = y + (gfx_sq_w - disc_diam)/2;
      
      if ((gfx_cont & GFX_BLACK_GHOST2) != 0) col = col_black; else col = col_white;

      g.setColor(col);
      // g.drawOval(xl, yt, disc_diam, disc_diam);
      Bresenham.drawCircle(g, xl, yt, disc_diam);
      
      int diam2 = (disc_diam*3)/4;
      diam2 &= ~1;
      
      xl = x + (gfx_sq_w - diam2)/2;
      yt = y + (gfx_sq_w - diam2)/2;
      // g.drawOval(xl, yt, diam2, diam2);
      Bresenham.drawCircle(g, xl, yt, diam2);
    }

    if ((gfx_cont & GFX_DISC_MASK) != 0) {

      // disc

      Color d_col;
      Color b_col;

      if ((gfx_cont & GFX_BLACK_DISC) != 0) {

        d_col = col_black;
        b_col = col_white;

      } else {

        b_col = col_black;
        d_col = col_white;

      }

      int xl = x + (gfx_sq_w - disc_diam)/2;
      int yt = y + (gfx_sq_w - disc_diam)/2;
      g.setColor(d_col);
      //g.fillOval(xl, yt, disc_diam, disc_diam);
      Bresenham.fillCircle(g, xl, yt, disc_diam);

      g.setColor(b_col);
      // g.drawOval(xl, yt, disc_diam, disc_diam);
      Bresenham.drawCircle(g, xl, yt, disc_diam);
    }

    if ((gfx_cont & GFX_BLACK_MOVE) != 0 ||
        (gfx_cont & GFX_WHITE_MOVE) != 0) {

      // legal move: draw X
      
      int cx = x + gfx_sq_w/2;
      int cy = y + gfx_sq_w/2;
      int l = gfx_sq_w/16;
      
      if ((gfx_cont & GFX_BLACK_MOVE) != 0)
        g.setColor(col_black);
      else 
        g.setColor(col_white);
      
      g.drawLine(cx-l,   cy-l, cx+l,   cy+l);
      g.drawLine(cx-l-1, cy-l, cx+l-1, cy+l);
      g.drawLine(cx-l,   cy+l, cx+l,   cy-l);
      g.drawLine(cx-l-1, cy+l, cx+l-1, cy-l);
      
    }

    if ((gfx_cont & GFX_LAST_MOVE) != 0) {

      // last move: draw inner square
      
      Color b_col;

      if ((gfx_cont & GFX_BLACK_DISC) != 0 ||
          (gfx_cont & GFX_HIT) != 0) {

        b_col = col_white;

      } else {

        b_col = col_black;

      }

      g.setColor(b_col);

      int dot_diam = disc_diam/4;
      dot_diam &= ~1;
      
      int xl = x + (gfx_sq_w - dot_diam)/2;
      int yt = y + (gfx_sq_w - dot_diam)/2;
      g.drawRect(xl, yt, dot_diam, dot_diam);

    }
  }

  protected void init_special_border(MailBoxGameState state) { }

  public String info_string() {
    
    int e = cont_num((AmazonsGameState)current_state(), EMPTY);
    int a = WIDTH*HEIGHT;
    
    return "(" + (a-e) + " " + (e) + ")";
  }

  protected String read_extra_info(MailBoxGameState st, StringStream is) {
    return null;
  }
  
  protected String extra_info_to_string(MailBoxGameState st) {
    return "";
  }
  
  public boolean paint_all(Graphics g, boolean repaint,
			   int gfx_sq[], int prev_gfx_sq[],
			   boolean xflip, boolean yflip) {
    return false;
  }  

}

