/* $Id: Ataxx.java,v 1.10 2006/08/30 08:52:09 orts_mburo Exp $
   
   (c) Michael Buro, licensed under the GPL2
*/

package ggs.ggsa._ataxx;

import java.util.*;
import java.awt.*;
import java.math.*;
import ggs.ggsa.boardgamesvc.*;
import ggs.ggsa.main.*;
import ggs.shared.*;

public class Ataxx extends BoardGameServiceClient
{
  public Ataxx() { super(new AtaxxGame()); }
  
  public void setDefaultOptions() 
  {
    new AtaxxGame().setDefaultOptions();
  }

}


class AtaxxGame extends MailBoxBoardGame
{
  public String get_name()    { return "Ataxx"; }  
  public String get_base_name()    { return "ax"; }
  public BoardGame new_BoardGame() { return new AtaxxGame(); }

  public void setDefaultOptions() {

    String sn = get_base_name() + ".";

    super.set_default_options();
    
    Global.options.put2i(sn+"ipw", BGS+"ipw"); // info panel
    Global.options.put2i(sn+"tph", BGS+"tph"); // game window vertical split

    Global.options.put2i(sn+"ge1", BGS+"ge1"); // 3 board window geoms
    Global.options.put2i(sn+"ge2", BGS+"ge2");
    Global.options.put2i(sn+"ge3", BGS+"ge3"); // also synchro geometry

    // playfield and grid color
    Global.options.put2i(sn+"gc",  BGS+"gc");  // grid color

    Global.options.put(sn+"mo",  new Boolean(true));         // show moves (boolean)
    Global.options.put(sn+"la",  new Boolean(true));         // show last move (boolean)

    // empty square color
    Global.options.put(sn+"es0", new Integer(15*65536 + 35*256 + 255));
    Global.options.put(sn+"es1", new Integer(15*65536 + 35*256 + 255));

    // blocked square color
    Global.options.put(sn+"bs",  new Integer(21*65536 + 28*256 + 163));

    // anti empty square color (not used)
    Global.options.put(sn+"aes0", new Integer(255*65536 + 114 * 256 + 86));
    Global.options.put(sn+"aes1", new Integer(255*65536 + 114 * 256 + 86));

    Global.options.put(sn+"cm", "");
  }
  
  public static final int BLOCKED = 4;

  public static final int GFX_BLACK_DISC   = 4;
  public static final int GFX_WHITE_DISC   = 8;
  public static final int GFX_BLOCKED      = 16;
  public static final int GFX_BLACK_MOVE   = 32;
  public static final int GFX_WHITE_MOVE   = 64;

  public static final int GFX_DISC_MASK = GFX_BLACK_DISC + GFX_WHITE_DISC;

  protected MailBoxGameState new_MailBoxGameState()
  {
    return new AtaxxGameState();
  }

  public boolean has_synchro()     { return true;  }
  public boolean has_komi()        { return true;  }
  public boolean has_rand()        { return true;  }
  public boolean has_rand_type()   { return true;  }
  public boolean has_anti()        { return false; }
  public boolean has_pref_color()  { return true;  }
  public boolean has_pass()        { return false; }

  public boolean never_turn()        { return true; }
  public boolean turned_if_black()   { return false; }
  protected boolean y_flip()         { return false; }
  protected boolean pieces_on_grid() { return false; }
  protected boolean only_squares()   { return true;  }

  protected int border_width()     { return 2; }

  public int board_type_num()      { return 11; }
  public int default_board_type()  { return 7; }

  public int board_type(int index)
  {
    if (index < 0 || index >= board_type_num()) return -1;
    return 4+index;
  }

  public int board_type_index(int type) {
    if (type < 4 || type > 14) return -1;
    return type-4;
  }

  public String board_type_string(int index) {
    int w = 4+index;
    return ""+w+"x"+w;
  }

  public int width(int type) { return type; }

  public int height(int type) { return width(type); }

  public byte char2cont(char c)
  {
    switch (c) {
    case '*': return BLACK;
    case 'O': return WHITE;
    case '-': return EMPTY;
    case '+': return BLOCKED;
    default:
      throw new Error("illegal char in Ataxx.char2cont");
    }
  }

  public char cont2char(byte co)
  {
    switch (co) {
    case BLACK: return '*';
    case WHITE: return 'O';
    case EMPTY: return '-';
    case BLOCKED:   return '+';

    default:
      throw new Error("illegal cont in Ataxx.cont2char");
    }
  }


  private boolean on_same_line(int ind1, int ind2)
  {
    int x1, y1, x2, y2;

    x1 = ind2x(ind1);
    y1 = ind2y(ind1);
    x2 = ind2x(ind2);
    y2 = ind2y(ind2);

    return x1 == x2 || y1 == y2 || Math.abs(x1-x2) == Math.abs(y1-y2);
  }


  private static int sign(int d)
  {
    if (d > 0) return 1;
    if (d < 0) return -1;
    return 0;
  }

  private int cont_num(AtaxxGameState state, int cont)
  {
    int n = 0;

    for (int i=0; i < BOARD_SIZE; ++i)
      if (state.sq[i] == cont) ++n;

    return n;
  }

  private boolean empty_line(AtaxxGameState st, int ind1, int ind2)
  {
    if (ind1 == ind2) return false;
    if (!on_same_line(ind1, ind2)) return false;
    int x1, y1, x2, y2;

    x1 = ind2x(ind1);
    y1 = ind2y(ind1);
    x2 = ind2x(ind2);
    y2 = ind2y(ind2);

    int dx = sign(x2-x1);
    int dy = sign(y2-y1);
    int d = dx + dy * DX;
    int p = ind1 + d;

    while (p != ind2) {
      if (st.sq[p] != EMPTY) return false;
      p += d;
    }

    return true;
  }


  private boolean move_at_ind(AtaxxGameState st, int ind, int color)
  {
    if (st.sq[ind] == color) {
      for (int dx=-2; dx <= 2; dx++) {
	for (int dy=-2; dy <= 2; dy++) {
	  if (st.sq[ind+dx+dy*DX] == EMPTY) return true;
	}
      }
    }
    return false;
  }


  private boolean no_moves(AtaxxGameState st, int color) {

    for (int i=0; i < BOARD_SIZE; ++i) {
      if (move_at_ind(st, i, color)) return false;
    }

    return true;
  }


  private boolean is_legal(AtaxxGameState state, int from, int to)
  {
    if (no_moves(state, state.to_move)) return false;

    if (from < 0 || from >= BOARD_SIZE || to < 0 || to >= BOARD_SIZE)
      return false;

    if (state.sq[from] != state.to_move) return false;
    if (state.sq[to] != EMPTY) return false;

    int x0 = ind2x(from);
    int y0 = ind2y(from);
    int x1 = ind2x(to);
    int y1 = ind2y(to);

    return Math.abs(x0-x1) <= 2 && Math.abs(y0-y1) <= 2;
  }


  public MailBoxGameState make_move2(String mv)
  {
    // parse move

    Vector<MovePart> v = new Vector<MovePart>();
    String err = parse_move(mv, v);

    if (err != null) return null;

    if (v.size() != 2) return null;

    MovePart mp = v.elementAt(0);
    if (!mp.rest.equals("")) return null;
    int from = mp.square;

    mp = v.elementAt(1);
    if (!mp.rest.equals("")) return null;
    int to = mp.square;

    // check move

    AtaxxGameState state = (AtaxxGameState) current_state();

    if (!is_legal(state, from, to)) return null;

    AtaxxGameState new_state = (AtaxxGameState) state.createCopy();

    new_state.sq[to] = (byte)new_state.to_move;

    int x0 = ind2x(from);
    int y0 = ind2y(from);
    int x1 = ind2x(to);
    int y1 = ind2y(to);

    if (Math.abs(x0-x1) >= 2 || Math.abs(y0-y1) >= 2) {
      // jump
      new_state.sq[from] = EMPTY;
    }

    byte opp_disc = (byte)PlayerColor.opponent(new_state.to_move);

    for (int i=0; i < 8; i++) {
      if (new_state.sq[to+D[i]] == opp_disc) {
	new_state.sq[to+D[i]] = (byte)new_state.to_move;
      }
    }

    if (!no_moves(new_state, PlayerColor.opponent(new_state.to_move))) {
      new_state.to_move = PlayerColor.opponent(new_state.to_move);
    }
    return new_state;
  }

  public void input(String s) { }

  protected void process_mouse_events() {

    boolean ok = false;
    int from = PASS;
    int to   = PASS;

    synchronized (mev) {

      int n = mev.size();
      int i;

      //       Global.dbgmsg("PROCESS_MOUSE_EVENTS BEFORE:");
      //       for (i=0; i < mev.size(); ++i) {
      // 	MouseEvent m = (MouseEvent)mev.elementAt(i);
      // 	Global.dbgmsg(m.type + " " + m.square);
      //       }

      for (i=0; i < mev.size(); ++i) {
	MouseEvent m = (MouseEvent)mev.elementAt(i);
	if (m.type == PRESSED) break;
	mev.removeElementAt(i);
	--i;
	continue;
      }

      ++i;

      for (;i < mev.size(); ++i) {
	MouseEvent m = (MouseEvent)mev.elementAt(i);
	if (m.type == RELEASED) break;
	mev.removeElementAt(i);
	--i;
	continue;
      }

      ++i;

      //       Global.dbgmsg("PROCESS_MOUSE_EVENTS AFTER:");

      //       for (i=0; i < mev.size(); ++i) {
      // 	MouseEvent m = (MouseEvent)mev.elementAt(i);
      // 	Global.dbgmsg(m.type + " " + m.square);
      //       }

      if (mev.size() == 0) return;

      AtaxxGameState state = (AtaxxGameState)current_state();

      if (mev.size() >= 1) {
	from = ((MouseEvent)mev.elementAt(0)).square;
	if (state.sq[from] != state.to_move) { reset_mev(); return; }
      }

      if (mev.size() >= 2) {
	to = ((MouseEvent)mev.elementAt(1)).square;
	if (state.sq[to] != EMPTY) { reset_mev(); return; }
	ok = true;
      }
    }

    if (ok) {
      reset_mev();
      handler.handle_event(new EventMsg(GameWindow.USER_MOVE_ID,
					ind2string(from) + "-" +
					ind2string(to)));
    }
  }

  protected boolean entire_gfx_cont(MailBoxGameState st, String last_move) {

    AtaxxGameState state = (AtaxxGameState)st;

    for (int i=0; i < BOARD_SIZE; i++) {

      if (state.sq[i] == BORDER) {

        gfx_sq[i] = GFX_BORDER;

      } else if (state.sq[i] == BLOCKED) {

        gfx_sq[i] = GFX_BLOCKED;

      } else {

        if (state.sq[i] == BLACK)
          gfx_sq[i] = GFX_BLACK_DISC;

        else if (state.sq[i] == WHITE)
          gfx_sq[i] = GFX_WHITE_DISC;

        else
          gfx_sq[i] = GFX_EMPTY;

      }
    }

    if (show_last_move) {
      Vector<MovePart> v = new Vector<MovePart>();
      if (parse_move(last_move, v) == null && v.size() == 2) {
	gfx_sq[v.elementAt(1).square] |= GFX_LAST_MOVE;
      }
    }

    int move_flag = ((state.to_move == PlayerColor.BLACK) ?
		     GFX_BLACK_MOVE :
		     GFX_WHITE_MOVE);
    int opp_move_flag = ((state.to_move == PlayerColor.BLACK) ?
			 GFX_WHITE_MOVE :
			 GFX_BLACK_MOVE);

    synchronized (mev) {

      if (mev.size() == 0) {

	if (show_moves) {

	  // no click: mark ataxx to move

	  for (int i=0; i < BOARD_SIZE; i++) {

	    if (state.sq[i] == state.to_move) {

	      // move possible?

	      if (move_at_ind(state, i, state.to_move)) {
		gfx_sq[i] |= opp_move_flag;
	      }
	    }
	  }
	}

      } else if (mev.size() == 1) {

	if (show_moves) {

	  int from = ((MouseEvent)mev.elementAt(0)).square;

	  // one click: mark target squares

	  for (int dx=-2; dx <= 2; dx++) {
	    for (int dy=-2; dy <= 2; dy++) {
	      int to = from + dx + dy * DX;
	      if (is_legal(state, from, to)) {
		gfx_sq[to] |= move_flag;
	      }
	    }
	  }
	}

      }
    }
    return true;
  }

  protected int last_move_index(String move)
  {
    Vector<MovePart> v = new Vector<MovePart>();
    if (parse_move(move, v) != null) return -1;
    return v.elementAt(1).square;
  }

  public int gfx_cont(MailBoxGameState st, int i) { return GFX_EMPTY; }

  protected void paint_gfx_cont(Graphics g, int x, int y, int parity, int gfx_cont) {

    if ((gfx_cont & GFX_BORDER) != 0) return;

    if ((gfx_cont & GFX_BLOCKED) != 0) {

      g.setColor(new Color(Global.options.getInteger(get_base_name()+".bs").intValue()));

    } else {

      if (parity == 0) g.setColor(col_empty0); else g.setColor(col_empty1);

    }

    g.fillRect(x+1,y+1,gfx_sq_w-1, gfx_sq_h-1);
    int disc_diam = (int)Math.round(gfx_sq_w * 0.85);
    disc_diam &= ~1;

    if ((gfx_cont & GFX_DISC_MASK) != 0) {

      // disc

      Color d_col;
      Color b_col;

      if ((gfx_cont & GFX_BLACK_DISC) != 0) {

        d_col = col_black;
        b_col = col_white;

      } else {

        b_col = col_black;
        d_col = col_white;

      }

      int xl = x + (gfx_sq_w - disc_diam)/2;
      int yt = y + (gfx_sq_w - disc_diam)/2;
      g.setColor(d_col);
      //g.fillOval(xl, yt, disc_diam, disc_diam);
      Bresenham.fillCircle(g, xl, yt, disc_diam);

      g.setColor(b_col);
      // g.drawOval(xl, yt, disc_diam, disc_diam);
      Bresenham.drawCircle(g, xl, yt, disc_diam);
    }

    if ((gfx_cont & GFX_BLACK_MOVE) != 0 ||
        (gfx_cont & GFX_WHITE_MOVE) != 0) {

      // legal move: draw X

      int cx = x + gfx_sq_w/2;
      int cy = y + gfx_sq_w/2;
      int l = gfx_sq_w/16;

      if ((gfx_cont & GFX_BLACK_MOVE) != 0)
        g.setColor(col_black);
      else
        g.setColor(col_white);

      g.drawLine(cx-l,   cy-l, cx+l,   cy+l);
      g.drawLine(cx-l-1, cy-l, cx+l-1, cy+l);
      g.drawLine(cx-l,   cy+l, cx+l,   cy-l);
      g.drawLine(cx-l-1, cy+l, cx+l-1, cy-l);

    }

    if ((gfx_cont & GFX_LAST_MOVE) != 0) {

      // last move: draw inner square

      Color b_col;

      if ((gfx_cont & GFX_BLACK_DISC) != 0 ||
          (gfx_cont & GFX_BLOCKED) != 0) {

        b_col = col_white;

      } else {

        b_col = col_black;

      }

      g.setColor(b_col);

      int dot_diam = disc_diam/4;
      dot_diam &= ~1;

      int xl = x + (gfx_sq_w - dot_diam)/2;
      int yt = y + (gfx_sq_w - dot_diam)/2;
      g.drawRect(xl, yt, dot_diam, dot_diam);

    }
  }

  protected void init_special_border(MailBoxGameState state) { }

  public String info_string() {

    int bd = cont_num((AtaxxGameState)current_state(), BLACK);
    int wd = cont_num((AtaxxGameState)current_state(), WHITE);

    return "" + bd + ":" + wd + " (" + (bd+wd) + " " + (bd-wd) + ")";
  }

  protected String read_extra_info(MailBoxGameState st, StringStream is) {
    return null;
  }

  protected String extra_info_to_string(MailBoxGameState st) {
    return "";
  }

  public boolean paint_all(Graphics g, boolean repaint,
			   int gfx_sq[], int prev_gfx_sq[],
			   boolean xflip, boolean yflip) {
    return false;
  }

}

