/* $Id: Go.java,v 1.10 2006/08/30 08:52:09 orts_mburo Exp $
   
   (c) Michael Buro, licensed under the GPL2
*/

package ggs.ggsa._go;

import java.util.*;
import java.awt.*;
import ggs.ggsa.boardgamesvc.*;
import ggs.ggsa.main.*;
import ggs.shared.*;

public class Go extends BoardGameServiceClient
{
  public Go() { super(new GoGame()); }
  
  public void setDefaultOptions() 
  {
    new GoGame().setDefaultOptions();
  }

}


class GoGame extends MailBoxBoardGame
{
  public String get_name()    { return "Go"; }  
  public String get_base_name()    { return "go"; }
  public BoardGame new_BoardGame() { return new GoGame(); }

  public void setDefaultOptions() {

    String sn = get_base_name() + ".";
    
    super.set_default_options();
    
    Global.options.put2i(sn+"ipw", BGS+"ipw"); // info panel
    Global.options.put2i(sn+"tph", BGS+"tph"); // game window vertical split

    Global.options.put2i(sn+"ge1", BGS+"ge1"); // 3 board window geoms
    Global.options.put2i(sn+"ge2", BGS+"ge2");
    Global.options.put2i(sn+"ge3", BGS+"ge3"); // also synchro geometry

    // playfield and grid color
    Global.options.put2i(sn+"pfc", BGS+"pfc"); // playfield color
    Global.options.put2i(sn+"gc",  BGS+"gc");  // grid color

    Global.options.put(sn+"mo",  new Boolean(true));         // show moves (boolean)
    Global.options.put(sn+"la",  new Boolean(true));         // show last move (boolean)

    // regular empty square color
    Global.options.put(sn+"es0", new Integer(205*65536  + 133*256 + 63));
    Global.options.put(sn+"es1", new Integer(205*65536  + 133*256 + 63));

    // last move square color
    Global.options.put(sn+"lms",  new Integer(255*65536 + 0*256 + 0));

    // anti empty square color (not used)
    Global.options.put(sn+"aes0", new Integer(255*65536 + 114 * 256 + 86));
    Global.options.put(sn+"aes1", new Integer(255*65536 + 114 * 256 + 86));    

    Global.options.put(sn+"cm", "");
  }

  // square content
  protected static final byte CONT_MASK   = 7;
    
  // square flags
  protected static final byte TERRITORY_B = 8;
  protected static final byte TERRITORY_W = 16;
  protected static final byte TERRITORY_MASK = TERRITORY_B+TERRITORY_W;
  protected static final byte MARKED_BY_B = 32;
  protected static final byte MARKED_BY_W = 64;
  protected static final byte MARKED_MASK = MARKED_BY_B + MARKED_BY_W;
  protected static final byte KILL_FLAG   = -128;
    
  // gfx content
  protected static final int GFX_BLACK       = 4;
  protected static final int GFX_WHITE       = 8;
  protected static final int GFX_TERRITORY_B = 16;
  protected static final int GFX_TERRITORY_W = 32;  
  
  protected static final int GFX_MARKED_BY_B = 64;
  protected static final int GFX_MARKED_BY_W = 128;
  
  protected static final int GFX_MOVE      = 256;
  protected static final int GFX_STONE_MASK = GFX_BLACK + GFX_WHITE;
  protected static final int GFX_TERRITORY_MASK =
    GFX_TERRITORY_B + GFX_TERRITORY_W;
  
  protected static final int GFX_MARKED_MASK = GFX_MARKED_BY_B + GFX_MARKED_BY_W;

  protected MailBoxGameState new_MailBoxGameState()
  {
    return new GoGameState();
  }

  public boolean has_synchro()     { return true;  }
  public boolean has_komi()        { return true;  }
  public boolean has_rand()        { return false; }
  public boolean has_rand_type()   { return false; }
  public boolean has_anti()        { return false; }  
  public boolean has_pref_color()  { return true;  }    
  public boolean has_pass()        { return true;  } 

  public boolean never_turn()        { return false; }  
  public boolean turned_if_black()   { return false; }
  protected boolean y_flip()         { return true; }
  protected boolean pieces_on_grid() { return true; }
  protected boolean only_squares()   { return true;  }

  protected int border_width()     { return 1; }

  public int board_type_num()      { return 9; }
  public int default_board_type()  { return 19; }

  public int board_type(int index)
  {
    return index*2+3;  // 3,5,7,...19
  } 

  public int board_type_index(int type) {

    int x = type / 100;
    int y = type % 100;

    if (x == 0) x = y;

    if (x != y) return -1;

    type = x;
    if (type < 3 || ((type - 3) & 1) == 1) return -1;
    int index = (type-3)/2;

    if (index >= board_type_num()) return -1;
    return index;
  }
  
  public String board_type_string(int index) {
    int x = index*2+3;
    return ""+x+"x"+x;
  }

  public int width(int type) { return type; }
  
  public int height(int type) { return width(type); }
  
  public byte char2cont(char c)
  {
    switch (c) {
    case '.': return EMPTY;   
    case 'X': return BLACK;  
    case 'O': return WHITE;
    default: throw new Error("illegal char in char2cont");
    }
  }

  // not used in scoring phase
  
  public char cont2char(byte cont)
  {
    switch (cont & CONT_MASK) {
    case EMPTY:       return '.';
    case BLACK:       return 'X';
    case WHITE:       return 'O';
    case TERRITORY_B: return '.';
    case TERRITORY_W: return '.';
    default: throw new Error("illegal cont in cont2char");
    }
  }

  public synchronized void input(String s) { }

  protected synchronized void process_mouse_events() {

    boolean ok = false;
    int square = -1;
    
    synchronized (mev) {
      
      int n = mev.size();

      for (MouseEvent m : mev) {
        if (m.type == CLICKED) {
          ok = true;
          square = m.square;
          reset_mev();
	  break;
        }
      }
    }

    if (ok) {
      handler.handle_event(new EventMsg(GameWindow.USER_MOVE_ID,
                                        ind2string(square)));
    }
  }

  protected synchronized boolean entire_gfx_cont(MailBoxGameState st0,
						 String last_move) {
    return false;
  }

  protected synchronized int last_move_index(String move)
  {
    Vector<MovePart> v = new Vector<MovePart>();
    if (parse_move(move, v) != null) return -1;
    return v.elementAt(0).square;
  }

  public int gfx_cont(MailBoxGameState st, int i)
  {
    int co = 0;

    switch (st.sq[i] & CONT_MASK) {
    case EMPTY:         co = GFX_EMPTY; break;
    case BLACK:         co = GFX_BLACK; break;
    case WHITE:         co = GFX_WHITE; break;
    default: throw new Error("unknown square contents");
    }

    if ((st.sq[i] & TERRITORY_MASK) == TERRITORY_B) co |= GFX_TERRITORY_B;
    if ((st.sq[i] & TERRITORY_MASK) == TERRITORY_W) co |= GFX_TERRITORY_W;

    if ((st.sq[i] & MARKED_BY_B) != 0) co |= GFX_MARKED_BY_B;
    if ((st.sq[i] & MARKED_BY_W) != 0) co |= GFX_MARKED_BY_W;    

    return co;
  }
  
  protected synchronized void paint_gfx_cont(Graphics g,
					     int x, int y, int parity, int gfx_cont)
  {
    final double STONE_FRAC     = 0.9;
    final double MARK_FRAC1     = 0.35;
    final double MARK_FRAC2     = 0.2;    
    final double TERRITORY_FRAC = 0.2;    

    if ((gfx_cont & GFX_BORDER) != 0) return;

    int gfx_w = gfx_sq_w & ~1;
    int gfx_h = gfx_sq_h & ~1;    
    int stone_diam = (int)Math.round(gfx_w * STONE_FRAC) & ~1;
    String sn = get_base_name() + ".";

    g.setColor(col_empty0);
    g.fillRect(x+1,y+1,gfx_sq_w, gfx_sq_h);

    // grid

    g.setColor(new Color(Global.options.getInteger(sn+"gc").intValue()));
    g.drawLine(x+gfx_sq_w/2, y, x+gfx_sq_w/2, y+gfx_sq_h);
    g.drawLine(x, y+gfx_sq_h/2, x+gfx_sq_w, y+gfx_sq_h/2);    

    if ((gfx_cont & GFX_STONE_MASK) != 0) {

      // draw stone

      Color d_col;
      Color b_col;

      if ((gfx_cont & GFX_BLACK) != 0) {
        d_col = col_black; b_col = col_white;
      } else {
        b_col = col_black; d_col = col_white;
      }

      int xl = x + (gfx_w - stone_diam)/2;
      int yt = y + (gfx_h - stone_diam)/2;

      g.setColor(d_col);
      Bresenham.fillCircle(g, xl, yt, stone_diam);

      // System.out.println(xl + " " + yt + " " + disc_diam + " ");
      
      g.setColor(b_col);
      Bresenham.drawCircle(g, xl, yt, stone_diam);

      if ((gfx_cont & GFX_MARKED_MASK) != 0) {

	// marked

	int disc_diam2 = (int)Math.round(gfx_w * MARK_FRAC1) & ~1;
	xl = x + (gfx_w - disc_diam2)/2;
	yt = y + (gfx_h - disc_diam2)/2;

	// remove circle
	
	g.setColor(col_empty0);
	Bresenham.fillCircle(g, xl, yt, disc_diam2);
	g.setColor(b_col);
	Bresenham.drawCircle(g, xl, yt, disc_diam2);
	
	// restore grid part

	g.setColor(col_black);
	g.drawLine(x+gfx_sq_w/2, y+gfx_sq_h/2-disc_diam2/2,
		   x+gfx_sq_w/2, y+gfx_sq_h/2+disc_diam2/2);
	
	g.drawLine(x+gfx_sq_w/2-disc_diam2/2, y+gfx_h/2,
		   x+gfx_sq_w/2+disc_diam2/2, y+gfx_sq_h/2);    
	

	// indicate marking color

	if ((gfx_cont & GFX_MARKED_BY_B) != 0) {
	  d_col = col_black; b_col = col_white;
	} else {
	  b_col = col_black; d_col = col_white;
	}

	int disc_diam3 = (int)Math.round(gfx_w * MARK_FRAC2) & ~1;
	xl = x + (gfx_w - disc_diam3)/2;
	yt = y + (gfx_h - disc_diam3)/2;
	
	g.setColor(d_col);
	Bresenham.fillCircle(g, xl, yt, disc_diam3);
	g.setColor(b_col);
	Bresenham.drawCircle(g, xl, yt, disc_diam3);
	
      }
      
    } else if ((gfx_cont & GFX_TERRITORY_MASK) != 0) {

      // draw territory point

      Color d_col;
      Color b_col;

      if ((gfx_cont & GFX_TERRITORY_B) != 0) {
        d_col = col_black; b_col = col_white;
      } else {
        b_col = col_black; d_col = col_white;
      }

      int disc_diam = (int)Math.round(gfx_w * TERRITORY_FRAC) & ~1;
      int xl = x + (gfx_w - disc_diam)/2;
      int yt = y + (gfx_h - disc_diam)/2;

      g.setColor(d_col);
      Bresenham.fillCircle(g, xl, yt, disc_diam);

      // System.out.println(xl + " " + yt + " " + disc_diam + " ");
      
      g.setColor(b_col);
      Bresenham.drawCircle(g, xl, yt, disc_diam);
    }

    if ((gfx_cont & GFX_LAST_MOVE) != 0) {

      g.setColor(new Color(Global.options.getInteger(sn+"lms").intValue()));

      int xl = x + (gfx_w - (stone_diam))/2;
      int yt = y + (gfx_h - (stone_diam))/2;
      Bresenham.drawCircle(g, xl, yt, stone_diam);
      xl = x + (gfx_w - (stone_diam+2))/2;
      yt = y + (gfx_h - (stone_diam+2))/2;
      Bresenham.drawCircle(g, xl, yt, stone_diam+2);
    }

  }

  protected void init_special_border(MailBoxGameState state) { }

  public synchronized String info_string() {

    GoGameState st = (GoGameState) current_state();

    if (st.scoring)
      return "P" + st.dead[WHITE] + "/" + st.dead[BLACK] +
	" T" + st.territory[BLACK] + "/" + st.territory[WHITE] + " " +
	(st.dead[WHITE]+st.territory[BLACK]-st.dead[BLACK]-st.territory[WHITE]);
    else
      return "ply " + (index+1) + " P" + st.dead[WHITE] + "/" + st.dead[BLACK];
  }

  public synchronized String read_extra_info(MailBoxGameState st0, StringStream is)
  {
    return null;
  }

  // weak test only
  
  public synchronized MailBoxGameState make_move2(String mv)
  {
    // parse move

    Vector<MovePart> v = new Vector<MovePart>();
    String err = parse_move(mv, v);

    if (err != null) return null;
    if (v.size() != 1) return null;

    MovePart mp = v.elementAt(0);
    if (!mp.rest.equals("")) return null;

    int move = mp.square;

    // check move

    GoGameState st = (GoGameState)(current_state().createCopy());

    if (st.scoring) {

      if (move != PASS) {

	int co = st.sq[move];

	if ((co & CONT_MASK) != BLACK && (co & CONT_MASK) != WHITE) {
	  Global.dbgmsg("scoring move on empty point");
	  return null;
	}

	int mark;
	
	if (st.to_move == BLACK) mark = MARKED_BY_B;
	else                     mark = MARKED_BY_W;

	if ((co & mark) != 0) {

	  // already marked: undo

	  mark_block(st, move, mark, false);

	} else {

	  // mark block
	  
	  mark_block(st, move, mark, true);

	  if ((st.sq[move] & MARKED_MASK) == MARKED_MASK) {

	    // both marks => kill block

	    erase(st, move, (co & CONT_MASK));
	  }
	}
      }
      
    } else {
    
      if (move == PASS) {

	MailBoxGameState pst = prev_state();
	if (pst != null && pst.move.toUpperCase().startsWith("PA")) {

	  // switch to scoring after 2 consecutive passes

	  st.scoring = true;
	}
	
      } else {
	
	if (st.sq[move] != EMPTY) return null;
	
	st.sq[move] = (byte)st.to_move;
	kill(st, move, PlayerColor.opponent(st.to_move));
      }
    }

    if (st.scoring) compute_territory(st);
    st.to_move = PlayerColor.opponent(st.to_move);
    return st;
  }


  private void mark_block(GoGameState st, int ind, int flags, boolean set)
  {
    int co = st.sq[ind] & CONT_MASK;
    if (co == BORDER) return;

    if (set) {
      if ((st.sq[ind] & flags) == flags) return; // already set
      st.sq[ind] |= flags;
    } else {
      if ((st.sq[ind] & flags) == 0) return; // already reset      
      st.sq[ind] &= ~flags;
    }

    for (int d = 0; d < 4; ++d) {
      int ind1 = ind + D4[d];
      if ((st.sq[ind1] & CONT_MASK) == co)
	mark_block(st, ind + D4[d], flags, set);
    }
  }

  
  private void erase(GoGameState st, int ind, int c)
  {
    ++st.dead[c]; st.sq[ind] = EMPTY;
    
    for (int d = 0; d < 4; ++d){
      int ind1 = ind + D4[d];
      if ((st.sq[ind1] & CONT_MASK) == c) erase(st, ind1, c);
    }
  }

  private boolean kill(GoGameState st, int ind, int c)
  {
    boolean any = false;
    
    for (int d = 0; d < 4; ++d) {
      int ind1 = ind + D4[d];
      if (st.sq[ind1] == c && is_dead(st, ind1, c)) {
	erase(st, ind1, c);
	any = true;
      }
    }
    
    return any;
  }

  boolean is_deadR(GoGameState st, int it, int c )
  {
    int dirset = 0;
    boolean dead = true;
    
    for (int d = 0; d < 4; ++d) { // note what dirs are left
      int cit = it + D4[d];
      if ((st.sq[cit] & KILL_FLAG) != 0) continue;
      if (st.sq[cit] == EMPTY) { dead = false; break; }
      if ((st.sq[cit] & ~KILL_FLAG) != c ) continue;
      dirset |= (1<<d);
      st.sq[cit] |= KILL_FLAG;
    }

    if (dead) {
      
      for ( int d = 0; d < 4; ++d ) { // go in each dir4
	if ((dirset & (1<<d)) != 0) {
	  if (!is_deadR(st, it + D4[d], c)) { // clean up when done
	    dead = false;
	    break;
	  }
	}
      }
    }

  CLEANUP:
    for (int d = 0; d < 4; ++d)
      if ((dirset & (1<<d)) != 0) st.sq[it+D4[d]] &= ~KILL_FLAG;
  
    return dead;
  }

  private boolean is_dead(GoGameState st, int it, int c )
  {
    st.sq[it] |= KILL_FLAG;
    boolean b = is_deadR(st, it, c );
    st.sq[it] &= ~KILL_FLAG;
    return b;
  }


  private void compute_territory(GoGameState st) {

    st.territory[0] = st.territory[1] = 0;

    // clear territory flags
    
    for (int i=0; i < BOARD_SIZE; ++i)
      if ((st.sq[i] & CONT_MASK) == EMPTY) {
	st.sq[i] &= ~TERRITORY_MASK;
      }

    // mark territories
    
    for (int i=0; i < BOARD_SIZE; ++i)
      if ((st.sq[i] & CONT_MASK) == BLACK)
	mark_territory(st, i, TERRITORY_B);
      else if ((st.sq[i] & CONT_MASK) == WHITE)
	mark_territory(st, i, TERRITORY_W);

    // count true territory
    
    for (int i=0; i < BOARD_SIZE; ++i)
      if ((st.sq[i] & TERRITORY_MASK) == TERRITORY_B)
	++st.territory[BLACK];
      else if ((st.sq[i] & TERRITORY_MASK) == TERRITORY_W)
	++st.territory[WHITE];
  }

  private void mark_territory(GoGameState st, int ind, int flags) {

    for (int d = 0; d < 4; ++d) {
      int ind1 = ind + D4[d];

      if ((st.sq[ind1] & CONT_MASK) == EMPTY &&
	  (st.sq[ind1] & flags) != flags) {
	mark_block(st, ind1, flags, true);
      }
    }
  }
  
  protected String extra_info_to_string(MailBoxGameState st) {
    return "";
  }
  
  public boolean paint_all(Graphics g, boolean repaint,
			   int gfx_sq[], int prev_gfx_sq[],
			   boolean xflip, boolean yflip) {
    return false;
  }  
  
}
