/* $Id: Hex.java,v 1.10 2006/08/30 08:52:09 orts_mburo Exp $
   
    (c) Michael Buro, licensed under the GPL2
*/

package ggs.ggsa._hex;

import java.util.*;
import java.awt.*;
import ggs.ggsa.boardgamesvc.*;
import ggs.ggsa.main.*;
import ggs.shared.*;

public class Hex extends BoardGameServiceClient
{
  public Hex() { super(new HexGame()); }
  
  public void setDefaultOptions() 
  {
    new HexGame().setDefaultOptions();
  }

}


class HexGame extends MailBoxBoardGame
{
  public String get_name()    { return "Hex"; }  
  public String get_base_name()    { return "hex"; }
  public BoardGame new_BoardGame() { return new HexGame(); }

  public void setDefaultOptions() {

    String sn = get_base_name() + ".";
    
    super.set_default_options();
    
    Global.options.put2i(sn+"ipw", BGS+"ipw"); // info panel
    Global.options.put2i(sn+"tph", BGS+"tph"); // game window vertical split

    Global.options.put2i(sn+"ge1", BGS+"ge1"); // 3 board window geoms
    Global.options.put2i(sn+"ge2", BGS+"ge2");
    Global.options.put2i(sn+"ge3", BGS+"ge3"); // also synchro geometry

    // playfield and grid color
    Global.options.put2i(sn+"gc",  BGS+"gc");  // grid color

    Global.options.put(sn+"mo",  new Boolean(true));         // show moves (boolean)
    Global.options.put(sn+"la",  new Boolean(true));         // show last move (boolean)

    // regular empty square color
    Global.options.put(sn+"es0", new Integer(135*65536 + 206 * 256 + 255));
    Global.options.put(sn+"es1", new Integer(135*65536 + 206 * 256 + 255));    

    // anti empty square color - not used
    Global.options.put(sn+"aes0", new Integer(255*65536 + 114 * 256 + 86));
    Global.options.put(sn+"aes1", new Integer(255*65536 + 114 * 256 + 86));    

    Global.options.put(sn+"cm", "");
  }


  // gfx content

  public static final int GFX_BLACK_DISC = 4;
  public static final int GFX_WHITE_DISC = 8;
  public static final int GFX_DISC_MASK  = GFX_BLACK_DISC + GFX_WHITE_DISC;

  private final double F  = 0.8660254; // w/h = sqrt(3)/2
  private int dx, zi, x0z, y0z;
  
  protected MailBoxGameState new_MailBoxGameState()
  {
    return new MailBoxGameState();
  }

  public boolean has_synchro()     { return true; }
  public boolean has_komi()        { return true; }
  public boolean has_rand()        { return true; }
  public boolean has_rand_type()   { return true; }
  public boolean has_anti()        { return false; }  
  public boolean has_pref_color()  { return true; }    
  public boolean has_pass()        { return false; } 

  public boolean never_turn()      { return false; }  
  public boolean turned_if_black() { return false; }
  protected boolean y_flip()       { return false; }

  public int board_type_num()      { return 8; }
  public int default_board_type()  { return 9; }

  public int board_type(int index)
  {
    switch (index) {
    case 0:  return 3;
    case 1:  return 5;
    case 2:  return 7;
    case 3:  return 9;
    case 4:  return 11;      
    case 5:  return 13;
    case 6:  return 15;
    case 7:  return 17;       
    default: return -1;
    }
  }

  public int board_type_index(int type) {
    switch (type) {
    case 3:  return 0;
    case 5:  return 1;
    case 7:  return 2;
    case 9:  return 3;
    case 11: return 4;      
    case 13: return 5;
    case 15: return 6;
    case 17: return 7;       
    default: return -1;
    }
  }
  
  public String board_type_string(int index) {
    switch (index) {
    case 0:  return "3x3";
    case 1:  return "5x5";
    case 2:  return "7x7";
    case 3:  return "9x9";
    case 4:  return "11x11";      
    case 5:  return "13x13";
    case 6:  return "15x15";
    case 7:  return "17x17";       
    default: return "?";
    }
  }

  public int width(int type) { return type; }
  
  public int height(int type) { return width(type); }
  
  public int border_width() { return 1; }

  public byte char2cont(char c)
  {
    switch (c) {
    case 'X': return BLACK;
    case 'O': return WHITE;
    case '.': return EMPTY;
    default:
      throw new Error("illegal char in Hex.char2cont");
    }
  }

  public char cont2char(byte co)
  {
    switch (co) {
    case BLACK: return 'X';
    case WHITE: return 'O';
    case EMPTY: return '.';
    default:
      throw new Error("illegal cont in Hex.cont2char");
    }
  }
  
  public synchronized void input(String s) { }

  private boolean is_legal(MailBoxGameState state, int ind)
  {
    return state.sq[ind] == EMPTY;
  }

  private int cont_num(MailBoxGameState state, int cont)
  {
    int n = 0;

    for (int i=0; i < BOARD_SIZE; ++i)
      if (state.sq[i] == cont)
	++n;

    return n;
  }
		       
  public MailBoxGameState make_move2(String mv)
  {
    // parse move

    Vector<MovePart> v = new Vector<MovePart>();
    String err = parse_move(mv, v);

    if (err != null) return null;
    if (v.size() != 1) return null;
    
    MovePart mp = v.elementAt(0);

    if (!mp.rest.equals("")) return null;

    int ind = mp.square;
    MailBoxGameState new_state = current_state().createCopy();

    if (!is_legal(new_state, ind)) return null;

    if (new_state.to_move == PlayerColor.BLACK) 
      new_state.sq[ind] = BLACK;
    else
      new_state.sq[ind] = WHITE;
    
    new_state.to_move = PlayerColor.opponent(new_state.to_move);
    return new_state;
  }


  public synchronized void mouse_action(String type, int mx, int my)
  {
    compute_sizes();

    if (mx < x0z || my < y0z) return;

    int y = (my - y0z)/gfx_sq_h;
    int x = (mx - x0z - y*dx)/gfx_sq_w;

    if (x < 0 || x >= WIDTH || y < 0 || y >= HEIGHT) return;

    int d1 = (my - y0z - gfx_sq_h/2 - gfx_sq_h * y);
    int d2 = (mx - x0z - gfx_sq_w/2 - y*dx - gfx_sq_w * x);

    //    Global.dbgmsg("" + d1*d1 + " " + d2*d2 + " " + (d1*d1 + d2*d2) + " " + gfx_sq_w*gfx_sq_w);
    
    if (d1*d1 + d2*d2 > gfx_sq_w*gfx_sq_w/9) return;

    if (y_flip() ^ turned) y = HEIGHT-1-y;
    if (turned) x = WIDTH-1-x;

    mev.addElement(new MouseEvent(type, xy2ind(x, y)));
    process_mouse_events();
    handler.handle_event(new EventMsg(GameWindow.REPAINT_ID));
  }

  protected void process_mouse_events()
  {
    boolean ok = false;
    int square = -1;
    
    synchronized (mev) {
      
      int n = mev.size();

      for (MouseEvent m : mev) {
	if (m.type == CLICKED) {
	  ok = true;
	  square = m.square;
	  reset_mev();
	  break;
	}
      }
    }

    if (ok) {
      handler.handle_event(new EventMsg(GameWindow.USER_MOVE_ID,
					ind2string(square)));
    }
  }

  protected int last_move_index(String move)
  {
    Vector<MovePart> v = new Vector<MovePart>();
    if (parse_move(move, v) != null) return -1;
    return v.elementAt(0).square;
  }

  public int gfx_cont(MailBoxGameState st, int i)
  {
    MailBoxGameState state = (MailBoxGameState)st;

    byte cont = state.sq[i];
    int gfx;
      
    if (cont == BORDER) return GFX_BORDER;

    if (cont == BLACK) 
      gfx = GFX_BLACK_DISC;
    else if (cont == WHITE)
      gfx = GFX_WHITE_DISC;
    else {
      gfx = GFX_EMPTY;
    }

    return gfx;
  }

  protected boolean entire_gfx_cont(MailBoxGameState st, String last_move) {
    return false;
  }


  private void hexagon(Graphics g, boolean fill,
		       int x0, int y0, int ix, int iy,
		       int zi, int dx,
		       int sq_w, int sq_h) {
    
    int x = x0+ix*sq_w + iy*dx;
    int y = y0+iy*sq_h;
    int xa[] = new int[6];
    int ya[] = new int[6];    
    xa[0] = x;        ya[0] = y+zi;
    xa[1] = x+sq_w/2; ya[1] = y-zi;
    xa[2] = x+sq_w;   ya[2] = y+zi;
    xa[3] = x+sq_w;   ya[3] = y+sq_h-zi;
    xa[4] = x+sq_w/2; ya[4] = y+sq_h+zi;
    xa[5] = x;        ya[5] = y+sq_h-zi;

    if (fill) g.fillPolygon(xa, ya, 6); else g.drawPolygon(xa, ya, 6); 
  }

  private void compute_sizes() {
    
    // WIDTH = HEIGHT

    int w1 = (int)(gfx_w/(WIDTH + (WIDTH-1)*0.5));
    int w2 = (int)(gfx_h/(WIDTH + F/4)/F);

    gfx_sq_w = Math.min(w1, w2);
    gfx_sq_h = (int)(gfx_sq_w * F);
    dx = (int)(gfx_sq_w*0.5);
    
    zi = (int)((((float)gfx_sq_w*gfx_sq_w)/gfx_sq_h)/8);
    x0z = x0;
    y0z = y0 + zi;
  }
  
  public boolean paint_all(Graphics g, boolean repaint,
			   int gfx_sq[], int prev_gfx_sq[],
			   boolean xflip, boolean yflip) {
    

    compute_sizes();
    
      
    Color col_grid = new Color(Global.options.getInteger(get_base_name()+"."+"gc").
			       intValue());
    
    if (repaint) {

      // draw hex grid

      Color col_char = Global.main_window.getForeground();
      
      for (int i=0; i < WIDTH; i++) {

	char xc;
	
	if (xflip)
	  xc = (char)('A'+WIDTH-1-i);
	else
	  xc = (char)('A'+i);
	
	g.setColor(col_char);
	g.drawString(String.valueOf(xc), 
		     x0 + i*gfx_sq_w + gfx_sq_w/2 - char_width/2, 
		     y0/2 + max_asc/2
		     );
      }
      
      for (int i=0; i < HEIGHT; i++) {

	int x = x0z;
	int ii = i;
	
	if (yflip) ii = HEIGHT-1-ii;
	if (ii > 8) x -= (3*char_width)/4;
	
	g.setColor(col_char);
	g.drawString(String.valueOf(ii+1), 
		     x - char_width - SPACE/2 + i*dx,
		     y0z + i*gfx_sq_h + gfx_sq_h/2 + max_asc - char_height/2
		     );
      }

      g.setColor(col_grid);
      
      for (int ix=0; ix < WIDTH; ix++) {
	for (int iy=0; iy < HEIGHT; iy++) {
	  hexagon(g, false, x0z, y0z, ix, iy, zi, dx, gfx_sq_w, gfx_sq_h);
	}
      }

      // black edge

      g.setColor(col_black);
      int ix, iy = WIDTH-1;
      
      for (ix=0; ix < WIDTH; ix++) {
	int x = x0z+ix*gfx_sq_w + iy*dx;
	int y = y0z+iy*gfx_sq_h + 1;
	g.drawLine(x, y+gfx_sq_h-zi, x+gfx_sq_w/2, y+gfx_sq_h+zi);
	g.drawLine(x+gfx_sq_w/2, y+gfx_sq_h+zi, x+gfx_sq_w, y+gfx_sq_h-zi);
	y++;
	g.drawLine(x, y+gfx_sq_h-zi, x+gfx_sq_w/2, y+gfx_sq_h+zi);
	g.drawLine(x+gfx_sq_w/2, y+gfx_sq_h+zi, x+gfx_sq_w, y+gfx_sq_h-zi);
      }


      // white edge

      g.setColor(col_white);
      ix = WIDTH-1;
      
      for (iy=0; iy < WIDTH; iy++) {
	int x = x0z+ix*gfx_sq_w + iy*dx + 1;
	int y = y0z+iy*gfx_sq_h;
	g.drawLine(x+gfx_sq_w/2, y-zi, x+gfx_sq_w, y+zi);
	g.drawLine(x+gfx_sq_w, y+zi, x+gfx_sq_w, y+gfx_sq_h-zi);
	x++;
	g.drawLine(x+gfx_sq_w/2, y-zi, x+gfx_sq_w, y+zi);
	g.drawLine(x+gfx_sq_w, y+zi, x+gfx_sq_w, y+gfx_sq_h-zi);
	x++;
	g.drawLine(x+gfx_sq_w/2, y-zi, x+gfx_sq_w, y+zi);
	g.drawLine(x+gfx_sq_w, y+zi, x+gfx_sq_w, y+gfx_sq_h-zi);
      }
    }
    
    for (int i=0; i < BOARD_SIZE; ++i) {

      if (repaint || gfx_sq[i] != prev_gfx_sq[i]) {

	int gfx_cont = gfx_sq[i];
	
	if ((gfx_cont & GFX_BORDER) == 0) {
	  int xi = ind2x(i);
	  int yi = ind2y(i);
	  if (y_flip() ^ turned) yi = HEIGHT-1-yi;
	  if (turned) xi = WIDTH-1-xi;
	  int x = x0z + xi * gfx_sq_w + yi * dx;
	  int y = y0z + yi * gfx_sq_h;

	  g.setColor(col_empty0);
	  hexagon(g, true, x0z, y0z, xi, yi, zi, dx, gfx_sq_w, gfx_sq_h);
	  g.setColor(col_grid);	  
 	  hexagon(g, false, x0z, y0z, xi, yi, zi, dx, gfx_sq_w, gfx_sq_h);
   
	  if ((gfx_cont & GFX_DISC_MASK) != 0) {

	    // disc

	    Color d_col;
	    Color b_col;

	    if ((gfx_cont & GFX_BLACK_DISC) != 0) {

	      d_col = col_black;
	      b_col = col_white;

	    } else {

	      b_col = col_black;
	      d_col = col_white;

	    }

	    int disc_diam = (int)Math.round(gfx_sq_w * 0.85) & ~1;
	    int xl = x + (gfx_sq_w - disc_diam)/2;
	    int yt = y + (gfx_sq_h - disc_diam)/2;

	    g.setColor(d_col);
	    Bresenham.fillCircle(g, xl, yt, disc_diam);

	    // System.out.println(xl + " " + yt + " " + disc_diam + " ");
      
	    g.setColor(b_col);
	    Bresenham.drawCircle(g, xl, yt, disc_diam);

	    if ((gfx_cont & GFX_LAST_MOVE) != 0) {

	      // last move: draw inner square

	      int dot_diam = disc_diam/8 & ~1;

	      xl = x + (gfx_sq_w - dot_diam)/2;
	      yt = y + (gfx_sq_h - dot_diam)/2;
	      g.drawRect(xl, yt, dot_diam, dot_diam);
	    }
	  }
	}
      }
    }
    
    return true;
  }  


  protected void init_special_border(MailBoxGameState state) {}

  public boolean only_squares() { return false; }

  public boolean pieces_on_grid() { return true; }
  
  public String info_string() {
    
    int bd = cont_num((MailBoxGameState)current_state(), BLACK);
    int wd = cont_num((MailBoxGameState)current_state(), WHITE);
    return "ply " + (bd+wd+1);
  }
  
  protected String extra_info_to_string(MailBoxGameState st) {
    return "";
  }

  public synchronized String read_extra_info(MailBoxGameState st0, StringStream is)
  {
    return null;
  }

  protected void paint_gfx_cont(Graphics g, int x, int y, int parity, int gfx_cont) {}
}

