/* $Id: Othello.java,v 1.9 2006/08/30 08:52:09 orts_mburo Exp $
   
   (c) Michael Buro, licensed under the GPL2
*/

package ggs.ggsa._othello;

import java.util.*;
import ggs.ggsa.boardgamesvc.*;
import ggs.ggsa.main.*;
import ggs.shared.*;

public class Othello extends BoardGameServiceClient
{
  public Othello() { super(new OthelloGame()); }
  
  public void setDefaultOptions() 
  {
    new OthelloGame().setDefaultOptions();
  }

}


class OthelloGame extends OthelloLikeBoardGame
{
  public String get_name()    { return "Othello"; }  
  public String get_base_name()    { return "os"; }
  public BoardGame new_BoardGame() { return new OthelloGame(); }

  public void setDefaultOptions()
  {
    String sn = get_base_name() + ".";

    super.set_default_options();
    
    Global.options.put2i(sn+"ipw", BGS+"ipw"); // info panel
    Global.options.put2i(sn+"tph", BGS+"tph"); // game window vertical split

    Global.options.put2i(sn+"ge1", BGS+"ge1"); // 3 board window geoms
    Global.options.put2i(sn+"ge2", BGS+"ge2");
    Global.options.put2i(sn+"ge3", BGS+"ge3"); // also synchro geometry

    // playfield and grid color
    Global.options.put2i(sn+"pfc", BGS+"pfc"); // playfield color
    Global.options.put2i(sn+"gc",  BGS+"gc");  // grid color

    Global.options.put(sn+"mo",  new Boolean(true));         // show moves (boolean)
    Global.options.put(sn+"la",  new Boolean(true));         // show last move (boolean)

    // regular empty square color
    Global.options.put(sn+"es0", new Integer(136*65536 + 188 * 256 + 136));
    Global.options.put(sn+"es1", new Integer(136*65536 + 188 * 256 + 136));

    // anti empty square color
    Global.options.put(sn+"aes0", new Integer(255*65536 + 114 * 256 + 86));
    Global.options.put(sn+"aes1", new Integer(255*65536 + 114 * 256 + 86));

    Global.options.put(sn+"cm", "");
  }


  protected MailBoxGameState new_MailBoxGameState()
  {
    return new MailBoxGameState();
  }

  public boolean has_synchro()     { return true; }
  public boolean has_komi()        { return true; }
  public boolean has_rand()        { return true; }
  public boolean has_rand_type()   { return true; }
  public boolean has_anti()        { return true; }
  public boolean has_pref_color()  { return true; }
  public boolean has_pass()        { return true; }

  public boolean never_turn()      { return true; }
  public boolean turned_if_black() { return false; }
  protected boolean y_flip()       { return false; }

  public int board_type_num()      { return 7; }
  public int default_board_type()  { return 8; }

  public int board_type(int index)
  {
    switch (index) {
    case 0:  return 4;
    case 1:  return 6;
    case 2:  return 8;
    case 3:  return 10;
    case 4:  return 12;
    case 5:  return 14;
    case 6:  return 88;
    default: return -1;
    }
  }

  public int board_type_index(int type) {
    switch (type) {
    case 4:  return 0;
    case 6:  return 1;
    case 8:  return 2;
    case 10: return 3;
    case 12: return 4;
    case 14: return 5;
    case 88: return 6;
    default: return -1;
    }
  }

  public String board_type_string(int index) {
    switch (index) {
    case 0:  return "4x4";
    case 1:  return "6x6";
    case 2:  return "8x8";
    case 3:  return "10x10";
    case 4:  return "12x12";
    case 5:  return "14x14";
    case 6:  return "88oct";
    default: return "?";
    }
  }

  public int width(int type) {
    if (type == 88) return 10;
    return type;
  }

  public int height(int type) { return width(type); }

  private int scan(MailBoxGameState state, int i, int d)
  {
    int opp = PlayerColor.opponent(state.to_move);

    i += d;

    //System.out.println("i=" + i + " d=" + d + " opp=" + opp + " sq=" + sq[i]);

    if (state.sq[i] == opp) {
      while (state.sq[i] == opp) i += d;
      if (state.sq[i] == state.to_move) {
	//System.out.println("legal");
	return i;
      }
    }

    //System.out.println("illegal");

    return -1;
  }

  private void flip(MailBoxGameState state, int i, int d)
  {
    int opp = PlayerColor.opponent(state.to_move);
    int i0 = i;

    i += d;

    if (state.sq[i] == opp) {
      while (state.sq[i] == opp) i += d;
      if (state.sq[i] == state.to_move) {
	do {
	  i -= d;
	  state.sq[i] = (byte)state.to_move;
	} while (i != i0);
      }
    }
  }

  private boolean is_legal(MailBoxGameState state, int ind)
  {
    if (ind == PASS) {

      for (int i=0; i < BOARD_SIZE; ++i) {
	if (is_legal(state, i))
	  return false;
      }

      return true;
    }

    if (state.sq[ind] != EMPTY)
      return false;

    for (int i: D) {
      if (scan(state, ind, i) >= 0)
	return true;
    }

    return false;
  }

  private int cont_num(MailBoxGameState state, int cont)
  {
    int n = 0;

    for (int i=0; i < BOARD_SIZE; ++i)
      if (state.sq[i] == cont)
	++n;

    return n;
  }

  public MailBoxGameState make_move2(String mv)
  {
    // parse move

    Vector<MovePart> v = new Vector<MovePart>();
    String err = parse_move(mv, v);

    if (err != null) return null;
    if (v.size() != 1) return null;

    MovePart mp = v.elementAt(0);

    if (!mp.rest.equals("")) return null;

    int ind = mp.square;
    MailBoxGameState new_state = current_state().createCopy();

    if (!is_legal(new_state, ind)) return null;

    if (ind != PASS) {
      for (int i : D) 
	flip(new_state, ind, i);

      new_state.sq[ind] = (byte)new_state.to_move;
    }

    new_state.to_move = PlayerColor.opponent(new_state.to_move);
    return new_state;
  }

  public void input(String s) { }

  protected void process_mouse_events() 
  {
    boolean ok = false;
    int square = -1;

    synchronized (mev) {

      int n = mev.size();

      for (MouseEvent m : mev) {
	if (m.type == CLICKED) {
	  ok = true;
	  square = m.square;
	  reset_mev();
	  break;
	}
      }
    }

    if (ok) {
      handler.handle_event(new EventMsg(GameWindow.USER_MOVE_ID,
					ind2string(square)));
    }
  }

  protected int last_move_index(String move)
  {
    Vector<MovePart> v = new Vector<MovePart>();
    if (parse_move(move, v) != null)
      return -1;
    return v.elementAt(0).square;
  }

  public int gfx_cont(MailBoxGameState st, int i)
  {
    MailBoxGameState state = (MailBoxGameState)st;

    byte cont = state.sq[i];
    int gfx;

    if (cont == BORDER) return GFX_BORDER;

    if (cont == BLACK)
      gfx = GFX_BLACK_DISC;
    else if (cont == WHITE)
      gfx = GFX_WHITE_DISC;
    else {
      gfx = GFX_EMPTY;
      if (show_moves && is_legal(state, i)) {
        if (state.to_move == BLACK)
          gfx |= GFX_BLACK_MOVE;
        else
          gfx |= GFX_WHITE_MOVE;
      }
    }

    return gfx;
  }

  protected boolean entire_gfx_cont(MailBoxGameState st, String last_move) {
    return false;
  }

  protected void init_special_border(MailBoxGameState state) {

    if (BOARD_TYPE == 88) {

      state.sq[xy2ind(0,0)] = BORDER;
      state.sq[xy2ind(0,1)] = BORDER;
      state.sq[xy2ind(1,0)] = BORDER;

      state.sq[xy2ind(9,9)] = BORDER;
      state.sq[xy2ind(9,8)] = BORDER;
      state.sq[xy2ind(8,9)] = BORDER;

      state.sq[xy2ind(0,9)] = BORDER;
      state.sq[xy2ind(0,8)] = BORDER;
      state.sq[xy2ind(1,9)] = BORDER;

      state.sq[xy2ind(9,0)] = BORDER;
      state.sq[xy2ind(8,0)] = BORDER;
      state.sq[xy2ind(9,1)] = BORDER;

    }
  }

  public String info_string() {

    int bd = cont_num((MailBoxGameState)current_state(), BLACK);
    int wd = cont_num((MailBoxGameState)current_state(), WHITE);

    return "" + bd + ":" + wd + " (" + (bd+wd) + " " + (bd-wd) + ")";
  }
}

