// $Id: CardHoldingPanel.java,v 1.3 2006/09/04 03:40:42 orts_mburo Exp $

package ggs.ggsa._skat.gui;

import ggs.ggsa.main.Global;

import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.BorderLayout;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import javax.swing.JPanel;
import javax.swing.JLabel;
import javax.swing.JButton;

import java.util.Observer;
import java.util.Observable;
import java.util.Vector;

/**
 * A JPanel that holds the CardPanels
 * 
 * @author Jan Sch&auml;fer <j@nschaefer.net>
 */
public class CardHoldingPanel extends JPanel implements Observer
{

  /**
   * Creates a new instance of CardHoldingPanel
   * 
   * @param bitmaps
   *            The SkatGraphicsRepository that holds all images that are
   *            used
   * @param player
   *            The player ID
   * @param panelType
   *            The type of the CardHoldingPanel
   * @param maxCardCount
   *            Maximal number of cards the panel can hold
   */
  public CardHoldingPanel(SkatGraphicRepository bitmaps,
			  int player,
			  PanelTypes panelType,
			  int maxCardCount)
  {
    this.bitmaps = bitmaps;
    this.player = player;
    this.maxCardCount = maxCardCount;
    this.cardCount = 0;

    this.panelType = panelType;
    cardPanels = new Vector<CardPanel>();

    initComponents();
		
    // FIXME remove this for-loop
    for (int i = 0; i < maxCardCount; i++) {
      cardPanels.get(i).setCard(i/8, i%8);
    }
		
    repaint();
  }

  private void initComponents()
  {
    // all CardHoldingPanels have BorderLayout
    setLayout(new BorderLayout());
    removeAll();
    // These panels are only placeholders for the BorderLayout
    JPanel northPanel = new JPanel();
    JPanel eastPanel = new JPanel();
    JPanel southPanel = new JPanel();
    JPanel westPanel = new JPanel();
    // the cards lie on the center panel
    JPanel centerPanel = new JPanel();

    // let the skat table shining through
    setOpaque(false);

    if (panelType == PanelTypes.NETWORK_PLAYER_PANEL) {
		
      // no gap between the cards of the opponent panels
      ((FlowLayout) centerPanel.getLayout()).setHgap(0);
    }

    CardPanelMouseAdapter mouseAdapter = new CardPanelMouseAdapter();

    // create the card panels
    for (int i = 0; i < maxCardCount; i++) {

      CardPanel panel;

      // create a card
      if (panelType == PanelTypes.NETWORK_PLAYER_PANEL) {

	// show the back of the opponent cards
	panel = new CardPanel(this, bitmaps, true);
				
      } else {

	// show the face of all the other cards
	panel = new CardPanel(this, bitmaps, false);
      }

      // different sizes for card panels
      if (i != (maxCardCount - 1)
	  && panelType == PanelTypes.NETWORK_PLAYER_PANEL) {

	// some card panels for the opponent cards are smaller
	panel.setPreferredSize(new Dimension(25, 97));
      } else {

	// all the other card panels have the normal size
	panel.setPreferredSize(new Dimension(73, 97));
      }

      // let the skat table shining through if there is no card shown
      panel.setOpaque(false);

      if (panelType == PanelTypes.PLAYER_PANEL
	  || panelType == PanelTypes.SKAT_PANEL
	  || panelType == PanelTypes.TRICK_PANEL) {

	// the mouse adapter listens to user clicks
	panel.addMouseListener(mouseAdapter);
      }

      // add it to the Vector that holds all panels
      cardPanels.add(panel);

      // add the card to the centerPanel
      centerPanel.add(panel);
    }

    if (panelType == PanelTypes.PLAYER_PANEL
	|| panelType == PanelTypes.NETWORK_PLAYER_PANEL) {

      // Show player name on the player panels
      JPanel playerNamePanel = new JPanel();
      playerName = new JLabel(" ");
      playerNamePanel.add(playerName);
      playerNamePanel.setPreferredSize(new Dimension(150,
						     (int) playerNamePanel.getPreferredSize().getHeight()));

      if (panelType == PanelTypes.PLAYER_PANEL) {

	southPanel.add(playerNamePanel);
				
      } else if (panelType == PanelTypes.NETWORK_PLAYER_PANEL) {

	northPanel.add(playerNamePanel);
      }
			
    } else if (panelType == PanelTypes.SKAT_PANEL) {

      // OK-Button for starting game after looking into the skat
      JButton okButton = new JButton("OK");

      okButton.addActionListener(new ActionListener() {

	  public void actionPerformed(ActionEvent evt) {

	    System.out.println("action for OK-Button on SKAT_PANEL not implemented, yet!");

	    // TODO move this handling to SkatGame
	    //implement it
	  }
	});

      southPanel.add(okButton);
    }

    // put all panels together
    centerPanel.setOpaque(false);
    add(centerPanel, BorderLayout.CENTER);
    northPanel.setOpaque(false);
    add(northPanel, BorderLayout.NORTH);
    eastPanel.setOpaque(false);
    add(eastPanel, BorderLayout.EAST);
    southPanel.setOpaque(false);
    add(southPanel, BorderLayout.SOUTH);
    westPanel.setOpaque(false);
    add(westPanel, BorderLayout.WEST);
  }

  /**
   * Implementation of the Observer pattern
   * 
   * @param observ
   *            The Observable that is observed
   * @param obj
   *            Object that has changed in the Observable
   */
  public void update(Observable observ, Object obj)
  {
    Global.dbgmsg("UPDATE panel #" + panelType + " " + observ + ": " + obj + " has changed...");
		
    // TODO implement update()
  }

  private void addCard()
  {
    Global.dbgmsg("addCard()");
		
    // TODO implement addCard()
  }

  private void removeCard()
  {
    Global.dbgmsg("removeCard()");
		
    // TODO implement removeCard()
  }

  private void reorganizeCards()
  {
    Global.dbgmsg("reorganizeCards()");
		
    // TODO implement reorganizeCard()
  }

  /**
   * Get the type of the CardHoldingPanel
   * 
   * @return The type of the CardHoldingPanel
   */
  public PanelTypes getPanelType()
  {
    return panelType;
  }

  /**
   * Get the player ID
   * 
   * @return The player ID
   */
  public int getPlayerID()
  {
    return player;
  }

  /**
   * Flips all cards
   */
  public void flipCards()
  {
    for (int i = 0; i < cardPanels.size(); i++) {
      ((CardPanel) cardPanels.get(i)).flipCard();
    }
		
    repaint();
  }

  /**
   * Is executed when a card panel is clicked
   * 
   * @param suit The suit of the clicked card panel
   * @param value The value of the clicked card panel
   */
  public void cardPanelClicked(int suit, int value)
  {
    // TODO implement cardPanelClicked()
  }

  public void setPlayerName(String newName)
  {
    playerName.setText(newName);
  }
	
  /**
   * All panel types
   */
  public enum PanelTypes {
    EMPTY_PANEL, NETWORK_PLAYER_PANEL,
      PLAYER_PANEL, SKAT_PANEL, TRICK_PANEL
      }

  private PanelTypes panelType;
  private SkatGraphicRepository bitmaps;
  private int player;
  private int maxCardCount;
  private int cardCount;
  private Vector<CardPanel> cardPanels;
  private JLabel playerName;
}
