/*
 * $Id: BoardCanvas.java,v 1.6 2006/08/28 06:57:30 orts_mburo Exp $
 * 
 * (c) Michael Buro, licensed under the GPL2
 */

package ggs.ggsa.boardgamesvc;
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import ggs.ggsa.main.*;
import ggs.shared.*;

public class BoardCanvas extends javax.swing.JPanel
  implements MouseListener
{
  private final boolean DBG_MSG = false;

  String id;
  EventHandler handler;
  protected int x0, y0, sq_len, disc_diam;
  int prev_width = -1;
  int prev_height = -1;
  Image double_buffer = null;
  BoardGame bg;
  boolean show_moves, show_last_move, turned;
  String last_move = "PA";

  // click bug:  click event not triggered if release location != press location
  // workaround: remember press location and generate click event if
  //             release location is "sufficiently" close to it
  
  int last_press_x, last_press_y;
  final int close_dist = 6;

  public void BoardCanvas()
  {
    initComponents();
  }
  
  public void init(BoardGame bg, String id, EventHandler handler)
  {
    this.bg = bg;
    bg.set_handler(handler);
    this.id = id;
    this.handler = handler;
    addMouseListener(this);
  }

  public void new_board_game(BoardGame bg2)
  {
    bg2.set_handler(handler);
    bg = bg2;
  }

  /** update flags and request draw */
  
  public void update(boolean show_moves,
                     boolean show_last_move,
                     String  last_move,
                     boolean turned
                     )
  {
    this.show_moves = show_moves;
    this.show_last_move = show_last_move;
    this.last_move = last_move;
    this.turned = turned;
    repaint();
  }


  public void paint(Graphics g)
  {
    super.paint(g);
    
    int cw = getSize().width;
    int ch = getSize().height;

    if (bg == null) 
      return;

    bg.paint(this, g, cw, ch, show_moves, show_last_move, last_move, turned);
  }

  
  public void mousePressed(MouseEvent e)
  {
    if (Global.options.getInteger("dbg").intValue() != 0)
      Global.dbgmsg("Canvas MousePressed: x="+ e.getX() + " y=" + e.getY());

    last_press_x = e.getX();
    last_press_y = e.getY();    

    handle_mouse_event(BoardGame.PRESSED, e);
  }

  /*  
      public void mouseClicked(MouseEvent e)					
      {										
      if (Global.options.getInteger("dbg").intValue() != 0)			
      Global.dbgmsg("Canvas MouseClicked: x="+ e.getX() + " y=" + e.getY());	
      handle_mouse_event(BoardGame.CLICKED, e);					
      }
  */

  public void mouseClicked(MouseEvent e) {} // buggy: do nothing
  
  public void mouseReleased(MouseEvent e)
  {
    if (Global.options.getInteger("dbg").intValue() != 0)
      Global.dbgmsg("Canvas MouseReleased: x="+ e.getX() + " y=" + e.getY());

    handle_mouse_event(BoardGame.RELEASED, e);

    // generate click event if close enough
    
    if (Math.abs(e.getX() - last_press_x) <= close_dist &&
	Math.abs(e.getY() - last_press_y) <= close_dist) {
      
      if (Global.options.getInteger("dbg").intValue() != 0)
	Global.dbgmsg("Canvas MouseClicked: x="+ e.getX() + " y=" + e.getY());

      handle_mouse_event(BoardGame.CLICKED, e);
    }
  } 

  private void handle_mouse_event(String type, MouseEvent e)
  {
    handler.handle_event(new EventMsg(id, type, ""+e.getX(), ""+e.getY()));
  }
  
  public void mouseEntered(MouseEvent e) {}
  public void mouseExited(MouseEvent e) {}

  
  /** This method is called from within the constructor to
   * initialize the form.
   * WARNING: Do NOT modify this code. The content of this method is
   * always regenerated by the Form Editor.
   */
  // <editor-fold defaultstate="collapsed" desc=" Generated Code ">//GEN-BEGIN:initComponents
  private void initComponents() {

    org.jdesktop.layout.GroupLayout layout = new org.jdesktop.layout.GroupLayout(this);
    this.setLayout(layout);
    layout.setHorizontalGroup(
			      layout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
			      .add(0, 400, Short.MAX_VALUE)
			      );
    layout.setVerticalGroup(
			    layout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
			    .add(0, 300, Short.MAX_VALUE)
			    );
  }// </editor-fold>//GEN-END:initComponents
  
  
  // Variables declaration - do not modify//GEN-BEGIN:variables
  // End of variables declaration//GEN-END:variables
  
}
