/* $Id: EdgeBoardGame.java,v 1.7 2006/08/28 06:57:30 orts_mburo Exp $
   
   (c) Michael Buro, licensed under the GPL2
*/

package ggs.ggsa.boardgamesvc;

import java.awt.*;
import java.util.*;
import javax.swing.*;
import ggs.ggsa.main.*;
import ggs.shared.*;

public abstract class EdgeBoardGame extends BoardGame
{
  protected class MouseEvent
  {
    public String type;
    public int square;

    public MouseEvent(String t, int s) { type = t; square = s; }
  }

  protected class MovePart
  {
    public int    square; // coordinate
    public String rest;   // additional characters
  }

  // additional square contents
  public static final byte EMPTY  = 2;
  public static final byte BORDER = 3;
  public static final byte CONT_MASK = 0x3f;

  // square kind
  public static final byte NODE   = 0x00;  
  public static final byte FACE   = 0x40;
  public static final byte EDGE_H = -128; // 0x80
  public static final byte EDGE_V = -64;  // 0xc0
  public static final byte KIND_MASK = -64; // 0xc0
  
  // pass coordinate (indicating pass move) must be < 0
  public static final int  PASS = -1;   

  // global gfx content
  public static final int GFX_EMPTY      = 0;
  public static final int GFX_BORDER     = 1;
  public static final int GFX_LAST_MOVE  = 0x8000;  
  
  // dimensions (set by init_dims)
  protected int BOARD_TYPE;
  protected int WIDTH, HEIGHT, DX;
  protected int BOARD_SIZE;
  protected int N, W, S, E, NW, NE, SW, SE; // compass directions
  protected int D[] = new int[8];           // stored in array
  protected int D4[] = new int[4];          // stored in array  

  // game states
  protected Vector<EdgeGameState> states;   // EdgeGameState[]
  protected int index = 0;                  // move index
    
  // graphics data
  protected JComponent component;
  protected MediaTracker mtrack;
  protected boolean show_last_move, show_moves, turned; // state!
  protected int gfx_sq[];           // flags
  protected int x0, y0;             // upper left board corner 
  protected int gfx_w, gfx_h;       // total gfx width & height
  protected int gfx_sq_w, gfx_sq_h; // square dimensions
  protected int char_width, char_height, char_max, max_asc;
  protected final int SPACE = 10;
  protected Color col_black = new Color(0,0,0);
  protected Color col_white = new Color(255,255,255);
  protected Color col_empty0, col_empty1;
  protected Color col_aempty0, col_aempty1;  
  
  protected int prev_gfx_sq[];
  protected int prev_gfx_w, prev_gfx_h;
  protected boolean prev_show_last_move, prev_show_moves, prev_turned;

  // mouse event data

  protected Vector<MouseEvent> mev = new Vector<MouseEvent>();

  protected void reset_mev() { mev.removeAllElements(); }
  
  protected int xy2ind(int x, int y) {
    if (x < 0 || x >= WIDTH || y < 0 || y >= HEIGHT)
      Global.errmsg("xy2ind: x or y out of range");
    return (y+border_width()) * DX + x + border_width();
  }

  protected int ind2x(int ind) { return ind % DX - border_width(); }
  protected int ind2y(int ind) { return ind / DX - border_width(); }

  protected String ind2string(int ind) {
    int x = ind2x(ind);
    int y = ind2y(ind);
    return (char)('A'+x) + "" + (y + 1);
  }
  
  // returns the MovePart Vector representation of mv

  protected String parse_move(String mv, Vector<MovePart> v)
  {
    StringTokenizer st = new StringTokenizer(mv, "-", true);
    Vector<String> strings = new Vector<String>();
    
    while (st.hasMoreTokens()) {
      String s = st.nextToken();
      if (!s.equals("-")) strings.addElement(s);
    }

    v.removeAllElements();
    
    for (String t : strings) {

      String s = t.toUpperCase();

      MovePart mp = new MovePart();
      mp.rest = "";
      
      if (s.equals("%") || (has_pass() && (s.equals("PA") || s.equals("PASS")))) {

	mp.square = PASS;

      } else { 

	StringStream iss = new StringStream(s);
	char c;
      
	c = iss.get();
	if (!iss.ok()) return "illegal move-x";

	int x = Character.toUpperCase(c) - 'A';
	if (x < 0 || x >= WIDTH) return "move-x out of range";

	int y = iss.get_int();
	if (!iss.ok()) return "illegal move-y";
	--y;
	if (y < 0 || y >= HEIGHT) return "move-y out of range";

	mp.square = xy2ind(x, y);
      
	while (true) {  // collect remaining characters
	  c = iss.get();
	  if (iss.ok()) mp.rest += c; else break;
	}
      }

      v.addElement(mp);
    }

    return null;
  }

  protected boolean init_dims(int bt) {

    WIDTH = 2*width(bt) - 1;
    if (WIDTH < 0) return false;

    HEIGHT = 2*height(bt) - 1;
    if (HEIGHT < 0) return false;
    
    BOARD_TYPE = bt;

    DX = WIDTH + 2*border_width();

    N = -DX; S = DX; W = -1; E = 1;
    NW = N+W; NE = N+E; SW = S+W; SE = S+E;
    
    D[0] = N; D[1] = NE; D[2] = E; D[3] = SE;
    D[4] = S; D[5] = SW; D[6] = W; D[7] = NW;     

    D4[0] = N; D4[1] = E; D4[2] = S; D4[3] = W;    

    BOARD_SIZE = DX*(HEIGHT + 2*border_width());
    return true;
  }
  
  protected String to_ggf(boolean one_line)
  {
    EdgeGameState state = current_state();

    String s = BOARD_TYPE + " " + extra_info_to_string(state);

    if (!one_line) s += "\n" ; else s += " ";
    
    for (int y=0; y < HEIGHT; ++y) {
      String sy = "";
      for (int x=0; x < WIDTH; x++) {

	if ((x & 1) == 0 && (y & 1) == 0) sy += node_char(); // node
	else {
	  int ind = xy2ind(x,y);
	  byte cont = state.sq[ind];
	  if (cont != BORDER) {
	    if ((x & 1) == 1 && (y & 1) == 1) // square
	      sy += s_cont2char(cont);
	    else if ((y & 1) == 0)     // horizontal edge
	      sy += h_cont2char(cont);
	    else
	      sy += v_cont2char(cont); // vertical edge	      
	  }
	}
      }
      s += sy;
      if (one_line) s += " "; else s += "\n";
    }

    if (state.to_move == BLACK) s += "*"; else s += "O";
    return s;
  }

  protected EdgeGameState current_state() {
    return (EdgeGameState)states.elementAt(index);
  }

  protected EdgeGameState prev_state() {
    if (index == 0) return null;
    return (EdgeGameState)states.elementAt(index-1);
  }

  public String start_pos(String s)
  {
    char c;
    StringStream is = new StringStream(s);
    EdgeGameState state = new_EdgeGameState();

    int bt = is.get_int();
    if (!is.ok() || !init_dims(bt)) return "illegal board type";

    state.sq = new byte[BOARD_SIZE];

    for (int i=0; i < BOARD_SIZE; ++i) state.sq[i] = BORDER;
    for (int x=0; x < WIDTH; ++x)
      for (int y=0; y < HEIGHT; ++y) {

	byte kind;
	
	if ((x & 1) == 0 && (y & 1) == 0) kind = NODE;
	else if ((x & 1) == 1 && (y & 1) == 1) kind = FACE;
	else if ((y & 1) == 0) kind = EDGE_H;
	else kind = EDGE_V;

	state.sq[xy2ind(x,y)] = (byte)(EMPTY | kind);
      }

    init_special_border(state);
    
    String err = read_extra_info(state, is);
    if (err != null) return err;
    
    for (int y=0; y < HEIGHT; ++y) {
      for (int x=0; x < WIDTH; ++x) {
	int ind = xy2ind(x,y);
	if (state.sq[ind] != BORDER) { // allows for non-rectangular shapes
	  
	  do {
	    if (is.eos()) return "missing squares";
	    c = is.get_skip_ws();
	  } while (Character.isWhitespace(c));

	  byte cont;
	  
	  if ((x & 1) == 0 && (y & 1) == 0) {

	    if (c != node_char()) return node_char() + " expected (" + c + ")";
	    cont = EMPTY;

	  } else {

	    if ((x & 1) == 1 && (y & 1) == 1) // square
	      cont = s_char2cont(c);
	    else if ((y & 1) == 0)     // horizontal edge
	      cont = h_char2cont(c);
	    else
	      cont = v_char2cont(c);   // vertical edge
	  }

	  state.sq[ind] = (byte)(cont | (state.sq[ind] & KIND_MASK));
	}
      }
    }

    c = is.get_skip_ws();
    if (!is.ok()) return "missing color";

    c = Character.toUpperCase(c);

    if      (c == '*' || c == 'X') state.to_move = BLACK;
    else if (c == 'O') state.to_move = WHITE;
    else return "illegal color to move " + c;

    states = new Vector<EdgeGameState>();
    states.addElement(state);
    index = 0;
    return null;
  }

  public int to_move() { return current_state().to_move; }

  public boolean advance(int d) {
    reset_mev();

    index += d;
    boolean ok = index >= 0 && index < states.size();
    if (index < 0) index = 0;
    if (index >= states.size()) index = states.size()-1;
    return ok;
  }
  
  public boolean paint(JComponent component,
				    Graphics g, int iw, int ih,
				    boolean sms,
				    boolean slm,
				    String last_move,
				    boolean tur)
  {
    boolean altered = false;

    this.component = component;
    this.show_moves = sms;
    this.show_last_move = slm;
    this.turned = tur;

    mtrack = new MediaTracker(component);
    
    Font font = Global.core.font;
    FontMetrics fm = g.getFontMetrics();
      
    char_width  = fm.charWidth('A');
    char_height = font.getSize();
    char_max    = Math.max(char_width, char_height);
    max_asc     = fm.getMaxAscent();

    if (only_squares()) iw = ih = Math.min(iw, ih);

    gfx_w = iw - char_max - SPACE;
    gfx_h = ih - char_max - SPACE;

    boolean repaint = true; // was prev_gfx_sq == null;

    if (gfx_w != prev_gfx_w || gfx_h != prev_gfx_h)
      repaint = true;

    prev_gfx_w = gfx_w;
    prev_gfx_h = gfx_h;
    
    gfx_sq_w = gfx_w / WIDTH;
    gfx_sq_h = gfx_h / HEIGHT;

    x0 = char_max + SPACE/2;
    y0 = char_max + SPACE/2;

    if (turned != prev_turned) {

      // clear previous coordinates

      repaint = true;
      g.setColor(Global.main_window.getBackground());
      g.fillRect(0, 0, iw, ih);
    }

    String sn = get_base_name()+".";
    col_empty0=new Color(Global.options.getInteger(sn+"es0").intValue());
    col_empty1=new Color(Global.options.getInteger(sn+"es1").intValue());

    if (has_anti()) {
      col_aempty0=new Color(Global.options.getInteger(sn+"aes0").intValue());
      col_aempty1=new Color(Global.options.getInteger(sn+"aes0").intValue());
    }

    gfx_sq = new int[BOARD_SIZE];
    EdgeGameState state = current_state();

    // compute graphics for each square

    if (!entire_gfx_cont(state, last_move)) {  // do it yourself
      // or let it be done here ...
    
      int last_move_ind = last_move_index(last_move);
    
      for (int i=0; i < BOARD_SIZE; ++i) {

	int x = ind2x(i);
	int y = ind2y(i);
	
	if (state.sq[i] == BORDER)
	  gfx_sq[i] = GFX_BORDER;
	else
	  gfx_sq[i] = gfx_cont(state, i);
      }

      if (show_last_move && last_move_ind >= 0) {
	gfx_sq[last_move_ind] |= GFX_LAST_MOVE;
      }
    }

    // paint all

    if (paint_all(g, repaint, gfx_sq, prev_gfx_sq, turned, y_flip() ^ turned)) {

      altered = repaint;
      if (!altered) {
	for (int i=0; i < BOARD_SIZE; ++i) {
	  if (gfx_sq[i] != prev_gfx_sq[i]) { altered = true; break; }
	}
      }

    } else {
    
      if (repaint) {

	// draw grid

	boolean xf = turned;
	boolean yf = y_flip() ^ turned;
	Color col_char = Global.main_window.getForeground();
	Color col_grid = new Color(Global.options.getInteger(sn+"gc").intValue());      
      
	for (int i=0; i < WIDTH+1; i++) {

	  if (i < WIDTH) {

	    char xc;
	  
	    if (xf)
	      xc = (char)('A'+WIDTH-1-i);
	    else
	      xc = (char)('A'+i);
	  
	    g.setColor(col_char);
	    g.drawString(String.valueOf(xc), 
			 x0 + i*gfx_sq_w + gfx_sq_w/2 - char_width/2, 
			 y0/2 + max_asc/2
			 );
	  }

	  g.setColor(col_grid);
	  g.drawLine(x0+i*gfx_sq_w, y0, x0+i*gfx_sq_w, y0+HEIGHT*gfx_sq_h);
	}

	for (int i=0; i < HEIGHT+1; i++) {

	  if (i < HEIGHT) {
      
	    int x = x0;
	    int ii = i;
	  
	    if (yf) ii = HEIGHT-1-ii;
	    if (ii > 8) x -= (3*char_width)/4;
	  
	    g.setColor(col_char);
	    g.drawString(String.valueOf(ii+1), 
			 x - char_width - SPACE/2,
			 y0 + i*gfx_sq_h + gfx_sq_h/2 + max_asc - char_height/2
			 );
	  }
	
	  g.setColor(col_grid);
	  g.drawLine(x0, y0+i*gfx_sq_h, x0+WIDTH*gfx_sq_w, y0+i*gfx_sq_h);
	}
      }

      // paint squares
      
      for (int i=0; i < BOARD_SIZE; ++i) {

	if (repaint || gfx_sq[i] != prev_gfx_sq[i]) {
	  if (gfx_sq[i] != BORDER) {
	    altered = true;
	    int xi = ind2x(i);
	    int yi = ind2y(i);
	    if (y_flip() ^ turned) yi = HEIGHT-1-yi;
	    if (turned) xi = WIDTH-1-xi;
	    int x = x0 + xi * gfx_sq_w;
	    int y = y0 + yi * gfx_sq_h;
	    paint_gfx_cont(g, x, y, (xi ^ yi) & 1, gfx_sq[i]);
	  }
	}
      }
    }

    prev_gfx_sq = gfx_sq;
    prev_show_moves = show_moves;
    prev_show_last_move = show_last_move;
    prev_turned = turned;
    return altered;
  }

  public void mouse_action(String type, int mx, int my)
  {
    if (mx < x0 || my < y0) return;

    int x = (mx - x0)/gfx_sq_w;
    int y = (my - y0)/gfx_sq_h;

    if (x >= WIDTH || y >= HEIGHT) return;

    if (y_flip() ^ turned) y = HEIGHT-1-y;
    if (turned) x = WIDTH-1-x;

    mev.addElement(new MouseEvent(type, xy2ind(x, y)));
    process_mouse_events();
    handler.handle_event(new EventMsg(GameWindow.REPAINT_ID));
  }

  public String make_move(String move)
  {
    EdgeGameState new_state = make_move2(move);

    if (new_state == null) return "illegal move " + move ;

    current_state().move = move; // for displaying last move location
    
    ++index;

    if (index >= states.size()) {
      states.addElement(new_state);
    } else {
      states.setElementAt(new_state, index);
      states.setSize(index+1);
    }

    return null; // OK
  }

  protected abstract int  border_width();
  protected abstract int  width(int board_type);
  protected abstract int  height(int board_type);  
  protected abstract byte s_char2cont(char c);
  protected abstract char s_cont2char(byte co);
  protected abstract byte v_char2cont(char c);
  protected abstract char v_cont2char(byte co);
  protected abstract byte h_char2cont(char c);
  protected abstract char h_cont2char(byte co);
  protected abstract char node_char();
  protected abstract void init_special_border(EdgeGameState state);
  protected abstract boolean entire_gfx_cont(EdgeGameState st, String last_move);
  protected abstract boolean paint_all(Graphics g, boolean repaint,
				       int gfx_sq[], int prev_gfx_sq[],
				       boolean xflip, boolean yflip);
  
  protected abstract String read_extra_info(EdgeGameState st, StringStream is);
  protected abstract String extra_info_to_string(EdgeGameState st);
  
  protected abstract void paint_gfx_cont(Graphics g,
					 int gfx_x, int gfx_y, int parity, int gfx_cont);
  protected abstract int  gfx_cont(EdgeGameState state, int index);
  protected abstract boolean y_flip();
  protected abstract void process_mouse_events();
  protected abstract int last_move_index(String last_move);
  protected abstract boolean only_squares();
  protected abstract EdgeGameState make_move2(String move);

  protected abstract EdgeGameState new_EdgeGameState();

  public int get_index() { return index; }
}
