/* $Id: GameClock.java,v 1.2 2006/08/24 09:12:45 orts_mburo Exp $
   
   (c) Michael Buro, licensed under the GPL2
*/

package ggs.ggsa.boardgamesvc;

import java.util.*;
import ggs.ggsa.main.*;

public class GameClock implements Cloneable
{
  private boolean using_extension;
  
  private USEC now; // this is the current (left) time
  private USEC inc;
  private USEC ext;

  private USEC ini_set;
  private USEC inc_set;
  private USEC ext_set;

  private Date mark;
  private boolean ticking;
  

  public GameClock() { init(); }

  public void init()
  {
    ini_set = new USEC(15 * 60 * USEC.uSec);
    inc_set = new USEC(0);
    ext_set = new USEC(2 * 60 * USEC.uSec);
    reset();
  }

  public void reset()
  {                   // resets clock to settings
    using_extension = false;
    ticking = false;
    mark = new Date();
    now = ini_set.createCopy();
    inc = inc_set.createCopy();
    ext = ext_set.createCopy();
  }

  public GameClock(GameClock x) { this(); copyOf(x); }

  public GameClock createCopy() { return new GameClock(this); }
  
  public void copyOf(GameClock x)
  {
    using_extension = x.using_extension;
    now.copyOf(x.now);
    inc.copyOf(x.inc);
    ext.copyOf(x.ext);
    ini_set.copyOf(x.ini_set);
    inc_set.copyOf(x.inc_set);
    ext_set.copyOf(x.ext_set);    
  }

  // only starting time relevant
  
  public boolean equals(GameClock x)
  {
    if (x == null) return false;
    if (x == this) return true;

    if (!ini_set.equals(x.ini_set)) return false;
    if (!inc_set.equals(x.inc_set)) return false;
    if (!ext_set.equals(x.ext_set)) return false;
    return true;
  }

  
  public void start()
  {
    if (ticking) return;
    ticking = true;
    mark = new Date();
  }   


  public long stop(long elapsed) // if < 0 use wall clock
  {
    // if (!ticking ) return -1;   // no ticking in client mode

    if (elapsed < 0) elapsed = (new Date().getTime() - mark.getTime()) * 1000;

    ticking = false;
    now.uadd(inc.usec() - elapsed);
    
    if (now.usec() < 0 && !using_extension) {
      using_extension = true;
      now.uadd(ext.usec());
      ext.uset(0);
    }
    return elapsed;
  }

  public boolean running() { return ticking; }
  
  public boolean in_extension() { return using_extension; }

  public long time_left(boolean setting)
  {
    if (setting) return ini_set.usec();
    else         return now.usec();
  }


  public Vector<String> parse(String arg, String delim)
  {
    StringTokenizer st = new StringTokenizer(arg, delim, true);
    Vector<String> strings = new Vector<String>();    
    boolean prev_delim = true;

    while (st.hasMoreTokens()) {
      String s = st.nextToken();
      if (s.equals(delim)) {
	if (prev_delim) strings.addElement("");
      } else strings.addElement(s);
      prev_delim = s.equals(delim);
    }

    return strings;
  }

  
  public boolean fromString(String arg)    // this will read the settings!!
  {
    init();

    Vector<String> strings = parse(arg, "/");
    
    while (strings.size() < 3) strings.addElement("");
    String s1 = strings.get(0);
    String s2 = strings.get(1);
    String s3 = strings.get(2);

    Vector<String> strings1 = parse(s1, ",");
    Vector<String> strings2 = parse(s2, ",");
    Vector<String> strings3 = parse(s3, ",");    

    while (strings1.size() < 2) strings1.add("");
    while (strings2.size() < 2) strings2.add("");
    while (strings3.size() < 2) strings3.add("");

    s1 = strings1.get(0);
    s2 = strings2.get(0);
    s3 = strings3.get(0);    
    
    if (!s1.equals("") && !ini_set.fromString(s1)) return false;
    if (!s2.equals("") && !inc_set.fromString(s2)) return false;
    if (!s3.equals("") && !ext_set.fromString(s3)) return false;
    
    return true;
  }


  public String toString()
  {
    Global.dbgmsg("GameClock.toString: arg?");
    System.exit(20);
    return "foo";
  }
  
  public String toString(boolean setting)
  {
    String s;
    
    if (setting ) {
      
      s = new HHMMSS(ini_set.sec()).toString(5) + "/"
	+ new HHMMSS(inc_set.sec()).toString(5) + "/"
	+ new HHMMSS(ext_set.sec()).toString(5);
      
    } else {
      
      s = new HHMMSS(now.sec()).toString(5) + "/"
	+ new HHMMSS(inc.sec()).toString(5) + "/"
	+ new HHMMSS(ext.sec()).toString(5);

    }

    return s;
  }

}

