/*
 * $Id: GameWindow.java,v 1.20 2006/09/04 17:25:54 orts_mburo Exp $
 *
 * (c) Michael Buro, licensed under the GPL2
 */

package ggs.ggsa.boardgamesvc;
import ggs.ggsa.main.*;
import ggs.shared.*;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.lang.*;
import java.text.*;
import javax.swing.*;
import java.beans.*;


public class GameWindow
  extends javax.swing.JFrame
  implements ActionListener, ggs.shared.EventHandler, WindowListener
{
  private final boolean DBG_EVENTS = false;

  protected BoardGameServiceClient service;
  protected GameWindowProp prop;

  private Vector<String> observers = new Vector<String>();
  private int clock_running_for  = PlayerColor.UNDEF; // => none
  private float b_seconds, w_seconds;
  private Date last_move_time;
  private DateFormat df = new SimpleDateFormat("HH:mm:ss");
  private int  prev_clock_running_for = -1;

  private GGS_Move komi_move;

  private ClockEventFactory cl = null;

  private Integer handle_lock = new Integer(0);
  private boolean shutting_down = false;

  // from Clock
  public static final String CLOCK_ID           = "CLOCK";

  // from BoardGame/Canvas
  public static final String PLAYFIELD_ID       = "PLAYFIELD"; // mouse action
  public static final String USER_MOVE_ID       = "USERMOVE";
  public static final String REPAINT_ID         = "REPAINT";   // repaint request
  public static final String INPUT_ID           = "INPUT";     // input request

  public static final String QUIT_ID            = "QUIT";      // input request  

  public static final String FS_MOVE_ID         = "FS_MOVE";  // move from service
  public static final String TS_MOVE_ID         = "TS_MOVE";  // move to service

  public static final String SCORE_INPUT_PREFIX = "@ your offered score? ";
  public static final String KOMI_INPUT_PREFIX  = "@ your komi value? ";
  public static final String MOVE_ATTRIBUTE_PREFIX = "@ move attribute:";

  private static final boolean FLASHING = false; // flashing "?" (to move indicator)

  Vector<JCheckBoxMenuItem> size_menu_items = new Vector<JCheckBoxMenuItem>();
  private String login;
  
  /** Creates new form GameWindow */
  public GameWindow(BoardGameServiceClient service_, GameWindowProp prop_)
  {
    initComponents();

    setDefaultCloseOperation(DO_NOTHING_ON_CLOSE);
    
    this.service = service_;
    this.prop = prop_;

    board_canvas.init(prop.gg.get_board_game(), PLAYFIELD_ID, this);

    utime_label.setFont(Global.core.clock_font);
    ltime_label.setFont(Global.core.clock_font);    

    login = service.getLogin();
    
    // create menu bar

    JMenuBar menu_bar = new JMenuBar();
    JMenuItem mi;
    JCheckBoxMenuItem cbmi;

    // Window Menu

    JMenu window = menu_bar.add(new JMenu("Window"));
    JMenu game = menu_bar.add(new JMenu("Game"));
    
    // Show Moves

    window.add(cbmi = new JCheckBoxMenuItem("Show Moves", prop.show_moves));
    cbmi.setEnabled(true);
    cbmi.addItemListener(new ItemListener() {
        public void itemStateChanged(ItemEvent e) {
	  prop.show_moves = e.getStateChange() == ItemEvent.SELECTED;
	  update_all();
        }});


    // Last move

    window.add(cbmi = new JCheckBoxMenuItem("Last Move", prop.show_last_move));
    cbmi.setEnabled(true);
    cbmi.addItemListener(new ItemListener() {
        public void itemStateChanged(ItemEvent e) {
	  prop.show_last_move = e.getStateChange() == ItemEvent.SELECTED;
	  update_all();	    
        }});

    // turned board

    window.add(cbmi = new JCheckBoxMenuItem("Board Turned", prop.turned));
    cbmi.setEnabled(true);
    cbmi.addItemListener(new ItemListener() {
        public void itemStateChanged(ItemEvent e) {
	  prop.turned = e.getStateChange() == ItemEvent.SELECTED;
	  update_all();	    
        }});

    window.addSeparator();

    // Sizes

    for (int i=0; i < prop.wp.get_geom_num(); i++) {

      window.add(cbmi = new JCheckBoxMenuItem("Size " + (i+1)));

      size_menu_items.add(cbmi);
      
      if (i == prop.wp.get_index())
	cbmi.setState(true);

      final int i0 = i;
      
      cbmi.addItemListener(new ItemListener() {
	  public void itemStateChanged(ItemEvent e) {
	    if (e.getStateChange() == ItemEvent.SELECTED) {
	      prop.wp.set_index(i0);
	      set_size(prop.wp.get_geom().get_w(), prop.wp.get_geom().get_h());
	      // reset other checkboxes
	      for (int i=0; i < prop.wp.get_geom_num(); ++i)
		if (i != i0)
		  size_menu_items.get(i).setSelected(false);
	      update_all();
	    }
	  }});
    }

    // print dimensions

    mi = new JMenuItem("Dimensions");
    mi.setEnabled(true);
    mi.addActionListener(new ActionListener() {
        public void actionPerformed(ActionEvent e) {
	  text.append("w=" + getWidth() + " h=" + getHeight() +
		      " hs=" + (getHeight()-hsplit.getLastDividerLocation()) +
		      " vs=" + (getWidth()-vsplit.getLastDividerLocation()) + "\n");
        }});

    window.add(mi);
    
    // Set Size

    //setsize_menuitem = mi = (MenuItem) window.add(new MenuItem("Set Size"));
    //mi.setEnabled(true);
    //mi.addActionListener(this);

    // Exit

    window.addSeparator();

    window.add(mi = new JMenuItem("Clear chat area"));
    mi.setEnabled(true);
    mi.addActionListener(new ActionListener() {
        public void actionPerformed(ActionEvent e) {
	  text.setText("");
	}});
    
    window.addSeparator();

    window.add(mi = new JMenuItem("Quit"));
    mi.setEnabled(true);
    mi.addActionListener(new ActionListener() {
        public void actionPerformed(ActionEvent e) {
	  service.handle_event(new EventMsg(QUIT_ID, prop.gg.id()));
	}
      });

    // Game Menu

    menu_bar.add(game);

    // Observers

    game.add(mi = new JMenuItem("Observers"));
    mi.setEnabled(true);
    mi.addActionListener(new ActionListener() {
        public void actionPerformed(ActionEvent e) {
	  if (prop.gg.id().startsWith("."))
	    show_observers();
        }});

	
    // GGF

    game.add(mi = new JMenuItem("Print GGF"));
    mi.setEnabled(true);
    mi.addActionListener(new ActionListener() {
        public void actionPerformed(ActionEvent e) {
	  Global.main_window.append("\n"+prop.gg.to_ggf(false));
	}
      });

    game.addSeparator();

    // Adjourn

    game.add(mi = new JMenuItem("Adjourn"));
    mi.setEnabled(true);
    mi.addActionListener(new ActionListener() {
        public void actionPerformed(ActionEvent e) {
	  if (prop.gg.id().startsWith("."))
	    Global.core.send_to_server(login, "t " + login +
				       " break " + prop.gg.id());
        }});

    // Abort

    game.add(mi = new JMenuItem("Abort Request"));
    mi.setEnabled(true);
    mi.addActionListener(new ActionListener() {
        public void actionPerformed(ActionEvent e) {
	  if (prop.gg.id().startsWith("."))
	    Global.core.send_to_server(login, "t " + login +
				       " abort " + prop.gg.id());
	}
      });

    // Undo

    game.add(mi = new JMenuItem("Undo Request"));
    mi.setEnabled(true);
    mi.addActionListener(new ActionListener() {
        public void actionPerformed(ActionEvent e) {
	  if (prop.gg.id().startsWith("."))
	    Global.core.send_to_server(login, "t " + login +
				       " undo " + prop.gg.id());
	}
      });

    // Score

    game.add(mi = new JMenuItem("Score Offer..."));
    mi.setEnabled(true);
    mi.addActionListener(new ActionListener() {
        public void actionPerformed(ActionEvent e) {
	  if (prop.gg.id().startsWith("."))
	    set_input(SCORE_INPUT_PREFIX);
	}
      });

    // Resign

    game.add(mi = new JMenuItem("Resign"));
    mi.setEnabled(true);
    mi.addActionListener(new ActionListener() {
        public void actionPerformed(ActionEvent e) {
	  if (prop.gg.id().startsWith("."))
	    Global.core.send_to_server(login, "t " + login +
				       " resign " + prop.gg.id());
        }});

    setJMenuBar(menu_bar);
    
    // menu bar done

    if (prop.wp.get_geom().get_x() > 0) {
      setLocation(prop.wp.get_geom().get_x(), prop.wp.get_geom().get_y());
    }

    int dh = 0;
    if (!Global.from_main)
      dh = 20; // compensate for extra applet text

    int w = prop.wp.get_geom().get_w();
    int h = prop.wp.get_geom().get_h() + dh;    
    set_size(w, h);

    //     hsplit.addPropertyChangeListener(new PropertyChangeListener() {
    //         public void propertyChange(PropertyChangeEvent evt) {
    // 	  JSplitPane s = (JSplitPane) evt.getSource();
    // 	  text.append("hsplit changed: " + evt.getNewValue() + " w=" + getWidth() + " h=" + getHeight() +
    // 		      " hs=" + (getWidth()-hsplit.getLastDividerLocation()) +
    // 		      " vs=" + (getHeight()-vsplit.getLastDividerLocation()) + "\n");
    //         }});

    //     vsplit.addPropertyChangeListener(new PropertyChangeListener() {
    //         public void propertyChange(PropertyChangeEvent evt) {
    // 	  JSplitPane s = (JSplitPane) evt.getSource();
    // 	  text.append("vsplit changed: " + evt.getNewValue() + " w=" + getWidth() + " h=" + getHeight() +
    // 		      " hs=" + (getWidth()-hsplit.getLastDividerLocation()) +
    // 		      " vs=" + (getHeight()-vsplit.getLastDividerLocation()) + "\n");
    //         }});
    
    setTitle(title());

    //    Global.dbgmsg("w: " + prop.wp.get_geom().get_w() +
    //		  " h: " + prop.wp.get_geom().get_h() +
    //		  " x: " + prop.wp.get_geom().get_x() +
    //		  " y: " + prop.wp.get_geom().get_y());
    //
    //    Global.dbgmsg("type: " + prop.gg.get_type().toString());


    pass_button.setVisible(prop.gg.has_pass());

    pass_button.addActionListener(new ActionListener() {
        public void actionPerformed(ActionEvent e) {
	  GGS_Move gm = new GGS_Move();
	  gm.set_move("PA");
	  gm.set_eval(0);
	  move_accepted(gm);
	  update_all();
        }});

    prev_button.addActionListener(new ActionListener() {
        public void actionPerformed(ActionEvent e) {
	  if (prop.gg.advance(-1))
	    update_all();
        }});

    next_button.addActionListener(new ActionListener() {
        public void actionPerformed(ActionEvent e) {
	  if (prop.gg.advance(1))
	    update_all();	    
        }});

    begin_button.addActionListener(new ActionListener() {
        public void actionPerformed(ActionEvent e) {
	  prop.gg.advance(-Math.max(prop.gg.move_num()/5, 10));
	  update_all();
	}});

    end_button.addActionListener(new ActionListener() {
        public void actionPerformed(ActionEvent e) {
	  prop.gg.advance(Math.max(prop.gg.move_num()/5, 10));
	  update_all();
        }});

    input.addActionListener(this);

    // setResizable(true);

    addWindowListener(this);

    // compute window infos from game data

    adjust_window_clocks();

    //setBackground(new Color(Global.options.getInteger("bg").intValue()));
    //setForeground(new Color(Global.options.getInteger("fg").intValue()));

    // start ticker
    cl = new ClockEventFactory(CLOCK_ID, 0.98);
    cl.ef.add_handler(this);
    cl.start();

    kib_chooser.setSelectedIndex(1);
    
    setVisible(true);
    
    update_all();
  }


  public void actionPerformed(ActionEvent e) 
  {
    Object src = e.getSource();
    
    if (src == input) {
      
      if (prop.gg.id().startsWith(".")) {
	
	String s = input.getText();
	input.push();
	
	if (s == null || s.equals("")) {
	  Global.dbgmsg(">>> EMPTY STRING: " + s);
	  return;
	}

	if (s.charAt(0) == '@') {

	  if (s.startsWith(KOMI_INPUT_PREFIX)) {

	    float val;
	    String t = s.substring(KOMI_INPUT_PREFIX.length());
	    try { val = Float.valueOf(t).floatValue(); }
	    catch (NumberFormatException f) {
	      append("illegal komi value");
	      return;
	    }

	    handle_komi(val);
	    return;

	  } else if (s.startsWith(SCORE_INPUT_PREFIX)) {

	    float val;
	    String t = s.substring(SCORE_INPUT_PREFIX.length());
	    try { val = Float.valueOf(t).floatValue(); }
	    catch (NumberFormatException f) {
	      append("illegal score value");
	      return;
	    }

	    handle_score(val);
	    return;

	  } else if (s.startsWith(MOVE_ATTRIBUTE_PREFIX)) {

	    prop.gg.input(s.substring(MOVE_ATTRIBUTE_PREFIX.length()));
	    return;

	  } else {

	    append("illegal @ line");
	    return;
	  }
	}

	int index = kib_chooser.getSelectedIndex();

	switch (index) {

	case 0:

	  Global.core.send_to_server(login, "t " + login + " tell o " +
				     prop.gg.id() +
				     " " +
				     s);
	  append(df.format(new Date()) + " o: " + s);
	  break;

	case 1:
	  Global.core.send_to_server(login, "t " + login + " tell " +
				     prop.gg.id() +
				     " " +
				     s);
	  append(df.format(new Date()) + " a: " + s);
	  break;

	case 2:

	  Global.core.send_to_server(login, "t " + login + " tell p " +
				     prop.gg.id() +
				     " " +
				     s);
	  append(df.format(new Date()) + " p: " + s);
	  break;

	default:
	}
      }
    }
  }

  public String title()
  {
    return service.getName() + " " +
      prop.gg.black_name() + " vs " + prop.gg.white_name() + ": " + prop.gg.id();
  }

  public void new_ggs_game(GGS_Game gg)
  {
    prop.gg = gg;
    board_canvas.new_board_game(gg.get_board_game());
    adjust_window_clocks();
    prop.turned = prop.gg.to_be_turned(Global.options.getString("lo"));
    update_all();
  }

  public void windowClosing(WindowEvent event) {
    service.handle_event(new EventMsg(QUIT_ID, prop.gg.id()));
  }
  public void windowOpened(WindowEvent event ) { }
  public void windowIconified(WindowEvent event ) { }
  public void windowDeiconified(WindowEvent event ) { }
  public void windowClosed(WindowEvent event ) { }
  public void windowActivated(WindowEvent event ) { }
  public void windowDeactivated(WindowEvent event ) { }

  protected void update_game_type(GGS_Game gg, JLabel tf)
  {
    String msg = gg.type_to_string(gg.komi_valid(), gg.get_komi());

    if (!tf.getText().equals(msg))
      tf.setText(msg);
  }

  protected void update_last_move_info(JLabel u_label, JLabel l_label)
  {
    String b_last="", w_last="";

    JLabel b_label, w_label;

    if (black_up())  {
      b_label = u_label;
      w_label = l_label;
    } else {
      b_label = l_label;
      w_label = u_label;
    }

    MoveInfo omi = prop.gg.prev_info();

    if (omi != null) {
      if (omi.to_move == PlayerColor.BLACK) b_last = omi.move;
      else                                  w_last = omi.move;
    }

    omi = prop.gg.prev_prev_info();

    if (omi != null) {
      if (omi.to_move == PlayerColor.BLACK) b_last = omi.move;
      else                                  w_last = omi.move;
    }

    GGS_Move gm = new GGS_Move();
    String msg, cmsg;

    if (gm.fromString(b_last) == null)
      msg = gm.get_move() + " " + gm.get_eval();
    else
      msg = "   ";

    //cmsg = Global.center(msg, GameWindowComponents.max_move_text.length());
    if (!b_label.getText().equals(msg))
      b_label.setText(msg);

    if (gm.fromString(w_last) == null)
      msg = gm.get_move() + " " + gm.get_eval();
    else
      msg = "   ";
    //cmsg = Global.center(msg, GameWindowComponents.max_move_text.length());
    if (!w_label.getText().equals(msg))
      w_label.setText(msg);
  }

  protected void update_name(boolean to_move,
			     String name, float rating, JLabel tf)
  {
    String msg = (to_move) ? "> " : "  ";
    msg += name + " " + Math.round(rating);
    if (!tf.getText().equals(msg)) tf.setText(msg);
  }

  protected void update_menus() { }

  protected void update_clock(boolean black_clock,
			      boolean running,
			      JLabel tf
			      )
  {
    float seconds;

    if (running) {

      if (black_clock)
	b_seconds = (float)(prop.gg.current_info().b_clock.time_left(false) / USEC.uSec
			    - (new Date().getTime() - last_move_time.getTime())/1000.0);

      else

	w_seconds = (float)(prop.gg.current_info().w_clock.time_left(false) / USEC.uSec
			    - (new Date().getTime() - last_move_time.getTime())/1000.0);

    }

    if (!prop.gg.at_end()) {
      if (black_clock)
	b_seconds = (float)(prop.gg.current_info().b_clock.time_left(false) / USEC.uSec);
      else
	w_seconds = (float)(prop.gg.current_info().w_clock.time_left(false) / USEC.uSec);
    }

    if (black_clock) seconds = b_seconds; else seconds = w_seconds;

    double clock = Math.abs(seconds);
    int h = (int)(clock/3600);
    clock -= h * 3600;
    int m = (int)(clock/60);
    clock -= m * 60;
    int s = (int)clock;

    String cs = "";

    if (seconds >= 0) cs = "+"; else cs = "-";

    cs += h + ":";
    if (m < 10) cs += "0";
    cs += "" + m + ":";
    if (s < 10) cs += "0";
    cs += "" + s;

    Color fg, bg;

    if (black_clock) { bg = Color.black; fg = Color.white; }
    else             { bg = Color.white; fg = Color.black; }

    if (tf.getText().equals(cs) &&
	tf.getForeground().equals(fg) &&
	tf.getBackground().equals(bg))
      return;  // same state

    tf.setOpaque(true);
    tf.setText(cs);
    tf.setForeground(fg);
    tf.setBackground(bg);
    tf.repaint();
  }


  protected void update_status(JLabel tf)
  {
    String msg = prop.gg.info_string();

    // center text
    // String cmsg = Global.center(s, GameWindowComponents.max_status_text.length());

    if (!tf.getText().equals(msg))
      tf.setText(msg);
  }


  private boolean black_up() {
    return prop.turned ^ prop.gg.turned_if_black();
  }


  protected void update_clocks()
  {
    JLabel btime_label, wtime_label;
    
    if (black_up())  {
      btime_label = utime_label;
      wtime_label = ltime_label;
    } else {
      btime_label = ltime_label;
      wtime_label = utime_label;
    }
    
    update_clock(!prop.gg.komi_move() || (prop.gg.to_move() == PlayerColor.BLACK),
		 clock_running_for == PlayerColor.BLACK ||
		 clock_running_for == PlayerColor.ALL,
		 btime_label
		 );
    
    update_clock(prop.gg.komi_move() && (prop.gg.to_move() == PlayerColor.BLACK),
		 clock_running_for == PlayerColor.WHITE ||
		 clock_running_for == PlayerColor.ALL,
		 wtime_label
		 );
  }

  private void update_all()
  {

    // last move info

    update_last_move_info(uinfo_label, linfo_label);

    //       ||

    JLabel bname_label, wname_label;

    if (black_up())  {
      bname_label = uname_label;
      wname_label = lname_label;
    } else {
      bname_label = lname_label;
      wname_label = uname_label;
    }

    boolean to_move = (prop.gg.to_move() == PlayerColor.BLACK) ||
      (prop.gg.is_komi_game() && prop.gg.move_num() == 0);

    update_name(to_move, prop.gg.black_name(), prop.gg.black_rating(), bname_label);

    to_move = (prop.gg.to_move() == PlayerColor.WHITE) ||
      (prop.gg.is_komi_game() && prop.gg.move_num() == 0);

    update_name(to_move, prop.gg.white_name(), prop.gg.white_rating(), wname_label);


    // time

    update_clocks();

    // game type

    update_game_type(prop.gg, game_type_label);

    update_status(status_label);

    boolean b = prop.gg.advance(-1);
    if (b) prop.gg.advance(1);
    prev_button.setEnabled(b);
    begin_button.setEnabled(b);

    b = prop.gg.advance(1);
    if (b) prop.gg.advance(-1);
    next_button.setEnabled(b);
    end_button.setEnabled(b);

    update_canvas();
  }


  private void update_canvas() 
  {
    String last_move="PA";

    if (prop.show_last_move && prop.gg.prev_info() != null) {
      GGS_Move gm = new GGS_Move();
      if (gm.fromString(prop.gg.prev_info().move) == null)
	last_move = gm.move;
    }

    // board

    boolean show_moves = prop.show_moves;
    boolean show_last_move = prop.show_last_move;
    boolean turned = prop.turned;

    board_canvas.update(show_moves, show_last_move, last_move, turned);
  }


  private void adjust_window_clocks()
  {
    //    Global.dbgmsg("adjust_window_clocks: " + prop.gg.get_type().is_komi_game() +
    //	  " " + prop.gg.komi_valid);

    b_seconds = prop.gg.current_info().b_clock.time_left(false) / USEC.uSec;
    w_seconds = prop.gg.current_info().w_clock.time_left(false) / USEC.uSec;

    if (!prop.gg.id().startsWith("."))
      clock_running_for = PlayerColor.UNDEF;
    else {
      if (prop.gg.is_komi_game() && !prop.gg.komi_valid()) {
	clock_running_for = PlayerColor.ALL;
      } else {
	clock_running_for = prop.gg.to_move();
      }

    }
    last_move_time = new Date();
  }

  public boolean user_to_move(String login)
  {
    if (!prop.gg.is_komi_game() || prop.gg.komi_valid()) {

      if ((prop.gg.to_move() == PlayerColor.BLACK &&
	   !prop.gg.black_name().equals(login)) ||
	  (prop.gg.to_move() == PlayerColor.WHITE &&
	   !prop.gg.white_name().equals(login)))
	return false;
    }

    return true;
  }

  public boolean handle_event(EventMsg msg0)
  {
    if (shutting_down)
      return false;

    final EventMsg msg = msg0;
    
    if (DBG_EVENTS)
      Global.dbgmsg("BOARD EVENT: " + msg.get_sender() + " " + msg);
	  
    if (msg.get_sender().equals(CLOCK_ID)) {
	    
      // clock tick
	    
      if (clock_running_for == PlayerColor.UNDEF)
	return true;
	    
      update_clocks();
      return true;
	    
    } else if (msg.get_sender().equals(PLAYFIELD_ID)) {
	    
      // user playfield action
	    
      if (prop.gg.id().startsWith(".")) {
	      
	String login = Global.options.getString("lo");
	      
	if (!prop.gg.black_name().equals(login) &&
	    !prop.gg.white_name().equals(login))
	  return true;
	      
	if (!user_to_move(login)) {
	  // append("not your turn!");
	  return true;
	}
      }
	    
      EventArgEnum en = msg.get_enum();
      String type = en.next_arg();
      String tx = en.next_arg();
      String ty = en.next_arg();
      int gfx_x, gfx_y;
	    
      try { gfx_x = Integer.valueOf(tx).intValue(); }
      catch (NumberFormatException f) { throw new Error("no integer x"); }
	    
      try { gfx_y = Integer.valueOf(ty).intValue(); }
      catch (NumberFormatException f) { throw new Error("no integer y"); }
	    
      if (DBG_EVENTS)
	Global.dbgmsg("MOUSE: " + type + " " + gfx_x + " " + gfx_y);
	    
      prop.gg.mouse_action(type, gfx_x, gfx_y);
      return true;
	    
    } else if (msg.get_sender().equals(REPAINT_ID)) {
	    
      update_canvas();
	    
    } else if (msg.get_sender().equals(USER_MOVE_ID)) {
	    
      if (DBG_EVENTS)
	Global.dbgmsg("USER_MOVE: " + msg.get_enum().next_arg());
	    
      komi_move = new GGS_Move();
      komi_move.set_move(msg.get_enum().next_arg());
      komi_move.set_eval(0);
      komi_move.set_time(((float)((new Date().getTime()-
				   last_move_time.getTime())/1000.0)));
	    
      if (prop.gg.is_komi_game() && !prop.gg.komi_valid()) {
	      
	// komi move => ask for value
	      
	set_input(KOMI_INPUT_PREFIX);
	return true;
      }
	    
      if (!prop.gg.id().startsWith(BoardGameServiceClient.LOCAL_GAME)) {
	      
	// public game: send move to server
	      
	move_accepted(komi_move);
	      
      } else {
	      
	// local game: make move
	adjust_window_clocks();
	prop.gg.make_move(komi_move.toString());
	prop.gg.set_end();
	update_all();
      }
	    
      komi_move = null;
      return true;
	    
    } else if (msg.get_sender().equals(INPUT_ID)) {
	    
      if (DBG_EVENTS)
	Global.dbgmsg("INPUT: " + msg.get_enum().next_arg());

      set_input( MOVE_ATTRIBUTE_PREFIX + " " + msg.get_enum().next_arg());
      return true;
	    
    } else if (msg.get_sender().equals(FS_MOVE_ID)) {
	    
      // move from server
	    
      if (prop.editable) {
	      
	if (prop.gg.id().equals(BoardGameServiceClient.LOCAL_GAME)) {
		
	  Global.dbgmsg("server move for local game");
	  return true;
	}
	      
	prop.gg.advance(prop.gg.move_num());
	      
	EventArgEnum en = msg.get_enum();
	String t = en.next_arg();
	int prev_to_move = prop.gg.to_move();
	      
	if (DBG_EVENTS) Global.dbgmsg("server move string: " + t);
	      
	if (prop.gg.make_move(t) != null) {
	  Global.dbgmsg(">>> illegal move from server"); return true;
	}
	      
	adjust_window_clocks();
	update_all();
      }
	    
      return true;
	    
      //       } else if (msg.get_sender().equals(SET_SIZE_ID)) {
	    
      // 	prop.wp.set_geom(new Geometry(getSize().width,
      // 				      getSize().height,
      // 				      getLocation().x,
      // 				      getLocation().y
      // 				      ));
      // 	set_size(prop.wp.get_geom().get_w(), prop.wp.get_geom().get_h());
      // 	append("[ WINDOW: x=" + getLocation().x +
      // 	       " y=" + getLocation().y +
      // 	       " w=" + getSize().width +
      // 	       " h=" + getSize().height +" ]\n");

      // 	validate(); // was repaint
      // 	return true;

      //       } else if (msg.get_sender().equals(USE_SIZE_I_ID)) {

      // 	Global.dbgmsg("use size " + msg);

      // 	int i;

      // 	try { i = new Integer(msg.get_enum().next_arg()).intValue(); }
      // 	catch (NumberFormatException f) {
      // 	  throw new Error("no integer in size menu");
      // 	}

      // 	prop.wp.set_index(i-1);
      // 	set_size(prop.wp.get_geom().get_w(), prop.wp.get_geom().get_h());
      // 	validate(); // was repaint
      // 	return true;

      //       } else if (msg.get_sender().equals(SET_SIZE_I_ID)) {

      // 	Global.dbgmsg("set size " + msg);

      // 	int i;

      // 	try { i = new Integer(msg.get_enum().next_arg()).intValue(); }
      // 	catch (NumberFormatException f) {
      // 	  throw new Error("no integer in size menu");
      // 	}

      // 	prop.wp.set_index(i-1);
      // 	prop.wp.set_geom(new Geometry(getSize().width,
      // 				      getSize().height,
      // 				      getLocation().x,
      // 				      getLocation().y
      // 				      ));
      // 	set_size(prop.wp.get_geom().get_w(), prop.wp.get_geom().get_h());
      // 	validate(); // was repaint
      // 	append("[ WINDOW: x=" + getLocation().x +
      // 	       " y=" + getLocation().y +
      // 	       " w=" + getSize().width +
      // 	       " h=" + getSize().height);

      // 	return true;

    }

    return false;
  }

  protected void move_accepted(GGS_Move gm)
  {
    if (prop.gg.id().startsWith(".")) {

      if (prop.gg.is_komi_game() && !prop.gg.komi_valid()) {

	// komi move

	// stop player clock after komi move

	if (prop.gg.black_name().equals(Global.options.getString("lo")))
	  clock_running_for = PlayerColor.WHITE;
	else
	  clock_running_for = PlayerColor.BLACK;
      }

      gm.set_time(((float)((new Date().getTime()-
			    last_move_time.getTime())/1000.0)));
    }

    Global.core.send_to_server(service.getLogin(),
			       "t " + service.getLogin() +
			       " play " + prop.gg.id() + " " + gm.toString());
  }


  protected void handle_score(float val)
  {
    if (prop.gg.id().startsWith(".")) {

      Global.core.send_to_server(service.getLogin(),
				 "t "+service.getLogin()+" score "+prop.gg.id()+" "+val);
    }
  }

  protected void handle_komi(float val)
  {
    if (komi_move != null) {
      komi_move.set_eval(val);
      move_accepted(komi_move);
      komi_move = null;
    }
  }

  protected void ended(EventMsg msg)
  {
    prop.gg.set_id("end "+ prop.gg.id() + " " + service.nextEndId());
    setTitle(title());

    clock_running_for = PlayerColor.UNDEF;

    EventArgEnum evenum = msg.get_enum();

    while (evenum.has_more_args())
      append(evenum.next_arg());
    
    update_all();
  }

  protected void append(String s)
  {
    text.append("\n" + s);
    text.setCaretPosition(text.getDocument().getLength());
  }

  protected void set_input(String s)
  {
    input.setText(s);
    input.setCaretPosition(input.getDocument().getLength());
    input.requestFocus();
    Toolkit.getDefaultToolkit().beep();
  }

  public void shut_down()
  {
    shutting_down = true;

    // delete clock & dispose window

    Global.dbgmsg("game window shut down");
    cl.end(); cl = null;
    dispose();
  }

  public void arrived(String name)
  {
    append(df.format(new Date()) + " " + name + " is watching");

    for (String s : observers) {
      if (s.equals(name))
	return;
    }

    observers.addElement(name);
  }

  public void departed(String name)
  {
    append(df.format(new Date()) + " " + name + " left");

    for (String s : observers) {
      if (s.equals(name)) {
	observers.remove(s);
	return;
      }
    }

    return;
  }


  public void show_observers()
  {
    String s = "";

    Global.dbgmsg(""+observers.size());

    for (String t : observers) {
      s += t + " ";
    }
    append("observers: " + s);
  }


  private void set_size(int w, int h)
  {
    setSize(w, h);
    validate();
    
    int vs = h-prop.wp.get_vsplit();
    int hs = w-prop.wp.get_hsplit();

    //     Global.dbgmsg("hs=" + hs + " vs=" + hs);
    //     Global.dbgmsg("w =" + w + " h=" + h);
    //     Global.dbgmsg("hd=" + prop.wp.get_hsplit() + " vd=" + prop.wp.get_vsplit());        

    hsplit.setDividerLocation(hs);
    validate();
    vsplit.setDividerLocation(vs);
    validate();
    
    repaint();
  }

  // ---- below generated by netbeans ----
  
  /** This method is called from within the constructor to
   * initialize the form.
   * WARNING: Do NOT modify this code. The content of this method is
   * always regenerated by the Form Editor.
   */
  // <editor-fold defaultstate="collapsed" desc=" Generated Code ">//GEN-BEGIN:initComponents
  private void initComponents() {
    java.awt.GridBagConstraints gridBagConstraints;

    vsplit = new javax.swing.JSplitPane();
    talk_panel = new javax.swing.JPanel();
    jScrollPane1 = new javax.swing.JScrollPane();
    text = new javax.swing.JTextArea();
    input = new ggs.ggsa.main.JHistTextField();
    hsplit = new javax.swing.JSplitPane();
    info_panel = new javax.swing.JPanel();
    jPanel2 = new javax.swing.JPanel();
    uname_label = new javax.swing.JLabel();
    utime_label = new javax.swing.JLabel();
    uinfo_label = new javax.swing.JLabel();
    jPanel5 = new javax.swing.JPanel();
    lname_label = new javax.swing.JLabel();
    ltime_label = new javax.swing.JLabel();
    linfo_label = new javax.swing.JLabel();
    button_panel = new javax.swing.JPanel();
    kib_chooser = new javax.swing.JComboBox();
    end_button = new javax.swing.JButton();
    next_button = new javax.swing.JButton();
    prev_button = new javax.swing.JButton();
    begin_button = new javax.swing.JButton();
    game_type_label = new javax.swing.JLabel();
    status_label = new javax.swing.JLabel();
    pass_button = new javax.swing.JButton();
    board_canvas = new ggs.ggsa.boardgamesvc.BoardCanvas();
    jMenuBar1 = new javax.swing.JMenuBar();
    jMenu1 = new javax.swing.JMenu();

    setDefaultCloseOperation(javax.swing.WindowConstants.EXIT_ON_CLOSE);
    vsplit.setDividerLocation(450);
    vsplit.setDividerSize(6);
    vsplit.setOrientation(javax.swing.JSplitPane.VERTICAL_SPLIT);
    vsplit.setResizeWeight(1.0);
    text.setColumns(20);
    text.setEditable(false);
    text.setLineWrap(true);
    text.setRows(5);
    jScrollPane1.setViewportView(text);

    org.jdesktop.layout.GroupLayout talk_panelLayout = new org.jdesktop.layout.GroupLayout(talk_panel);
    talk_panel.setLayout(talk_panelLayout);
    talk_panelLayout.setHorizontalGroup(
					talk_panelLayout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
					.add(jScrollPane1, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, 601, Short.MAX_VALUE)
					.add(input, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, 601, Short.MAX_VALUE)
					);
    talk_panelLayout.setVerticalGroup(
				      talk_panelLayout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
				      .add(talk_panelLayout.createSequentialGroup()
					   .add(jScrollPane1, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, 116, Short.MAX_VALUE)
					   .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
					   .add(input, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, 30, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE))
				      );
    vsplit.setBottomComponent(talk_panel);

    hsplit.setDividerLocation(440);
    hsplit.setDividerSize(6);
    hsplit.setResizeWeight(1.0);
    jPanel2.setLayout(new java.awt.GridBagLayout());

    jPanel2.setBorder(javax.swing.BorderFactory.createBevelBorder(javax.swing.border.BevelBorder.RAISED));
    uname_label.setText("Name");
    jPanel2.add(uname_label, new java.awt.GridBagConstraints());

    utime_label.setText("Time");
    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 0;
    gridBagConstraints.gridy = 1;
    jPanel2.add(utime_label, gridBagConstraints);

    uinfo_label.setText("Info");
    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 0;
    gridBagConstraints.gridy = 2;
    jPanel2.add(uinfo_label, gridBagConstraints);

    jPanel5.setLayout(new java.awt.GridBagLayout());

    jPanel5.setBorder(new javax.swing.border.SoftBevelBorder(javax.swing.border.BevelBorder.RAISED));
    lname_label.setText("Name");
    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 1;
    gridBagConstraints.gridy = 1;
    jPanel5.add(lname_label, gridBagConstraints);

    ltime_label.setText("Time");
    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 1;
    gridBagConstraints.gridy = 2;
    jPanel5.add(ltime_label, gridBagConstraints);

    linfo_label.setText("Info");
    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 1;
    gridBagConstraints.gridy = 3;
    jPanel5.add(linfo_label, gridBagConstraints);

    button_panel.setLayout(new java.awt.GridBagLayout());

    kib_chooser.setModel(new javax.swing.DefaultComboBoxModel(new String[] { "to observers", "to all", "to players" }));
    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 0;
    gridBagConstraints.gridy = 5;
    gridBagConstraints.gridwidth = 4;
    gridBagConstraints.insets = new java.awt.Insets(4, 0, 4, 0);
    button_panel.add(kib_chooser, gridBagConstraints);

    end_button.setText(">>");
    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 2;
    gridBagConstraints.gridy = 4;
    button_panel.add(end_button, gridBagConstraints);

    next_button.setText(">");
    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 2;
    gridBagConstraints.gridy = 3;
    gridBagConstraints.fill = java.awt.GridBagConstraints.HORIZONTAL;
    button_panel.add(next_button, gridBagConstraints);

    prev_button.setText("<");
    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 1;
    gridBagConstraints.gridy = 3;
    gridBagConstraints.fill = java.awt.GridBagConstraints.HORIZONTAL;
    button_panel.add(prev_button, gridBagConstraints);

    begin_button.setText("<<");
    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 1;
    gridBagConstraints.gridy = 4;
    button_panel.add(begin_button, gridBagConstraints);

    game_type_label.setText("Game Type");
    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 1;
    gridBagConstraints.gridy = 0;
    gridBagConstraints.gridwidth = 2;
    gridBagConstraints.fill = java.awt.GridBagConstraints.VERTICAL;
    gridBagConstraints.ipady = 1;
    gridBagConstraints.insets = new java.awt.Insets(1, 0, 1, 0);
    button_panel.add(game_type_label, gridBagConstraints);

    status_label.setText("Status");
    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 1;
    gridBagConstraints.gridy = 1;
    gridBagConstraints.gridwidth = 2;
    button_panel.add(status_label, gridBagConstraints);

    pass_button.setText("pass");
    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 1;
    gridBagConstraints.gridy = 2;
    gridBagConstraints.gridwidth = 2;
    gridBagConstraints.insets = new java.awt.Insets(3, 0, 3, 0);
    button_panel.add(pass_button, gridBagConstraints);

    org.jdesktop.layout.GroupLayout info_panelLayout = new org.jdesktop.layout.GroupLayout(info_panel);
    info_panel.setLayout(info_panelLayout);
    info_panelLayout.setHorizontalGroup(
					info_panelLayout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
					.add(org.jdesktop.layout.GroupLayout.TRAILING, info_panelLayout.createSequentialGroup()
					     .add(info_panelLayout.createParallelGroup(org.jdesktop.layout.GroupLayout.TRAILING)
						  .add(button_panel, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, 154, Short.MAX_VALUE)
						  .add(org.jdesktop.layout.GroupLayout.LEADING, jPanel2, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, 154, Short.MAX_VALUE)
						  .add(org.jdesktop.layout.GroupLayout.LEADING, jPanel5, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, 154, Short.MAX_VALUE))
					     .add(0, 0, 0))
					);
    info_panelLayout.setVerticalGroup(
				      info_panelLayout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
				      .add(info_panelLayout.createSequentialGroup()
					   .add(jPanel5, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, 143, Short.MAX_VALUE)
					   .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
					   .add(jPanel2, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, 144, Short.MAX_VALUE)
					   .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
					   .add(button_panel, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, 148, Short.MAX_VALUE))
				      );
    hsplit.setRightComponent(info_panel);

    org.jdesktop.layout.GroupLayout board_canvasLayout = new org.jdesktop.layout.GroupLayout(board_canvas);
    board_canvas.setLayout(board_canvasLayout);
    board_canvasLayout.setHorizontalGroup(
					  board_canvasLayout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
					  .add(0, 439, Short.MAX_VALUE)
					  );
    board_canvasLayout.setVerticalGroup(
					board_canvasLayout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
					.add(0, 447, Short.MAX_VALUE)
					);
    hsplit.setLeftComponent(board_canvas);

    vsplit.setLeftComponent(hsplit);

    jMenu1.setText("Menu");
    jMenuBar1.add(jMenu1);

    setJMenuBar(jMenuBar1);

    org.jdesktop.layout.GroupLayout layout = new org.jdesktop.layout.GroupLayout(getContentPane());
    getContentPane().setLayout(layout);
    layout.setHorizontalGroup(
			      layout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
			      .add(vsplit, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, 603, Short.MAX_VALUE)
			      );
    layout.setVerticalGroup(
			    layout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
			    .add(org.jdesktop.layout.GroupLayout.TRAILING, vsplit, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, 609, Short.MAX_VALUE)
			    );
    pack();
  }// </editor-fold>//GEN-END:initComponents
  
  
  // Variables declaration - do not modify//GEN-BEGIN:variables
  private javax.swing.JButton begin_button;
  private ggs.ggsa.boardgamesvc.BoardCanvas board_canvas;
  private javax.swing.JPanel button_panel;
  private javax.swing.JButton end_button;
  private javax.swing.JLabel game_type_label;
  private javax.swing.JSplitPane hsplit;
  private javax.swing.JPanel info_panel;
  private ggs.ggsa.main.JHistTextField input;
  private javax.swing.JMenu jMenu1;
  private javax.swing.JMenuBar jMenuBar1;
  private javax.swing.JPanel jPanel2;
  private javax.swing.JPanel jPanel5;
  private javax.swing.JScrollPane jScrollPane1;
  private javax.swing.JComboBox kib_chooser;
  private javax.swing.JLabel linfo_label;
  private javax.swing.JLabel lname_label;
  private javax.swing.JLabel ltime_label;
  private javax.swing.JButton next_button;
  private javax.swing.JButton pass_button;
  private javax.swing.JButton prev_button;
  private javax.swing.JLabel status_label;
  private javax.swing.JPanel talk_panel;
  private javax.swing.JTextArea text;
  private javax.swing.JLabel uinfo_label;
  private javax.swing.JLabel uname_label;
  private javax.swing.JLabel utime_label;
  private javax.swing.JSplitPane vsplit;
  // End of variables declaration//GEN-END:variables
  
}
