/* $Id: ReqPopUp.java,v 1.7 2006/08/28 06:57:30 orts_mburo Exp $
   
   (c) Michael Buro, licensed under the GPL2
*/

package ggs.ggsa.boardgamesvc;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;

import ggs.ggsa.main.*;
import ggs.shared.*;

public class ReqPopUp extends JFrame
  implements ItemListener,ActionListener
{
  GGS_Game     gg;
  String       login;
  boolean      my_request;
  String       title0, title;

  EventHandler handler;
  Request      orig_req;

  JPanel     player_board_panel;
  
  JLabel     player_label;
  JTextField player_input;
  JLabel     board_label;
  JComboBox  board_choice;

  JPanel     mode_panel;

  JCheckBox  synchro_checkbox;
  JCheckBox  komi_checkbox;
  JCheckBox  anti_checkbox;
  JCheckBox  rand_checkbox;
  JTextField rand_type_input;

  JPanel     color_panel;
  JLabel     color_label;
  JCheckBox  black_checkbox;
  JCheckBox  white_checkbox;
  
  JPanel     time_panel;
  JLabel     time_label;
  JTextField time_input;

  JPanel     otime_panel;
  JLabel     otime_label;
  JTextField otime_input;

  JPanel     button_panel;

  JButton    submit_button;
  JButton    cancel_button;

  JButton    accept_button;
  JButton    decline_button;
  JButton    open0_button;

  // input state (default)
  
  int     board_type;

  String  game_id = ".!";
  boolean synchro = false;
  boolean komi  = false;
  boolean anti  = false;
  boolean rand  = false;
  boolean black = false;
  boolean white = false;

  String player = "";
  String time   = "15:00/00:00/02:00";
  String otime  = "same";
  String rand_type = "";

  boolean submit_ok = false;
  boolean accept_ok = false;

  public ReqPopUp(String login, EventHandler h, GGS_Game gg, boolean my_req)
  {
    handler = h;
    my_request = my_req;
    this.login = login;
    this.gg = gg;
    
    board_type = gg.default_board_type();

    Geometry geom = Global.options.getGeometry(BoardGame.BGS+"rge");
    if (geom.get_x() > 0) {
      setLocation(geom.get_x(), geom.get_y());
    }
    
    setSize(geom.get_w(), geom.get_h());

    title0 = "GGSA" + login + " Match Request";
    title = title0;
    setTitle(title);

    // setBackground(new Color(Global.options.getInteger("bg").intValue()));
    // setForeground(new Color(Global.options.getInteger("fg").intValue()));

    VertiLayout vl = new VertiLayout(VertiLayout.H_CENTER, 0);
    setLayout(vl);
    
    player_label = (my_request ? new JLabel("To") : new JLabel("From"));
    player_input = new JTextField(player, 15);
    board_label  = new JLabel("Board");
    board_choice = new JComboBox();

    for (int i=0; i < gg.board_type_num(); ++i) {
      board_choice.addItem(gg.board_type_string(i));
    }

    board_choice.addItemListener(this);
    
    player_board_panel = new JPanel();
    player_board_panel.setLayout(new FlowLayout(FlowLayout.CENTER, 0, 3));
    player_board_panel.add(player_label);
    player_board_panel.add(player_input);
    player_board_panel.add(board_label);
    player_board_panel.add(board_choice);

    add(player_board_panel);
    vl.setConstraints(player_board_panel,
		      new VertiLayoutConstraints(VertiLayoutConstraints.FIX));

    mode_panel = new JPanel();
    
    mode_panel.setLayout(new FlowLayout(FlowLayout.CENTER, 0, 3));

    if (gg.has_synchro()) {
      synchro_checkbox = new JCheckBox("Syn");
      mode_panel.add(synchro_checkbox);
      synchro_checkbox.addItemListener(this);
    }
    
    if (gg.has_komi()) {
      komi_checkbox = new JCheckBox("Komi");
      mode_panel.add(komi_checkbox);
      komi_checkbox.addItemListener(this);
    }
    
    if (gg.has_anti()) {
      anti_checkbox = new JCheckBox("Anti");
      mode_panel.add(anti_checkbox);      
      anti_checkbox.addItemListener(this);
    }

    if (gg.has_rand()) {
      rand_checkbox = new JCheckBox("Rand");
      mode_panel.add(rand_checkbox);

      rand_checkbox.addItemListener(this);

      if (gg.has_rand_type()) {
	rand_type_input = new JTextField(rand_type, 3);
	mode_panel.add(rand_type_input);

	rand_type_input.addActionListener(this);
      }
    }

    add(mode_panel);
    vl.setConstraints(mode_panel, new VertiLayoutConstraints(VertiLayoutConstraints.FIX));


    if (gg.has_pref_color()) {
      color_label    = new JLabel("Your Color");
      black_checkbox = new JCheckBox("Black"); 
      white_checkbox = new JCheckBox("White"); 
      color_panel = new JPanel();
      color_panel.setLayout(new FlowLayout(FlowLayout.CENTER, 0, 3));
      color_panel.add(color_label);
      color_panel.add(black_checkbox);
      color_panel.add(white_checkbox);
      add(color_panel);
      vl.setConstraints(color_panel,new VertiLayoutConstraints(VertiLayoutConstraints.FIX));

      black_checkbox.addItemListener(this);
      white_checkbox.addItemListener(this);
    }

    time_label = new JLabel("Your Time");
    time_input = new JTextField("", 17);
    time_panel = new JPanel();
    time_panel.setLayout(new FlowLayout(FlowLayout.CENTER, 0, 3));
    time_panel.add(time_label);
    time_panel.add(time_input);
    add(time_panel);
    vl.setConstraints(time_panel, new VertiLayoutConstraints(VertiLayoutConstraints.FIX));

     
    otime_label = new JLabel("Opps Time");
    otime_input = new JTextField("", 17);
    otime_panel = new JPanel();
    otime_panel.setLayout(new FlowLayout(FlowLayout.CENTER, 0, 3));
    otime_panel.add(otime_label);
    otime_panel.add(otime_input);
    add(otime_panel);
    vl.setConstraints(otime_panel, new VertiLayoutConstraints(VertiLayoutConstraints.FIX));
  
    button_panel = new JPanel();
    button_panel.setLayout(new FlowLayout(FlowLayout.CENTER, 5, 5));

    if (my_request) {
      
      submit_button = new JButton("Submit"); 
      cancel_button = new JButton("Cancel"); 
      
      button_panel.add(submit_button);
      button_panel.add(cancel_button);

      // button listeners
      
      submit_button.addActionListener(this);
      cancel_button.addActionListener(this);

    } else {
      
      accept_button  = new JButton("Accept");
      submit_button  = new JButton("Submit");
      decline_button = new JButton("Decline");
      open0_button   = new JButton("Open 0");
      button_panel.add(accept_button);
      button_panel.add(submit_button);
      button_panel.add(decline_button);
      button_panel.add(open0_button);

      // button listeners

      accept_button.addActionListener(this);
      decline_button.addActionListener(this);
      open0_button.addActionListener(this);
    }

    // input listeners
    
    player_input.addActionListener(this);
    time_input.addActionListener(this);
    otime_input.addActionListener(this);

    add(button_panel);    
    vl.setConstraints(button_panel,new VertiLayoutConstraints(VertiLayoutConstraints.FIX));

    set();
    setOK();
    validate();
    setVisible(false);
  }

  // false = error
  
  public boolean fromRequest(Request req)
  {
    if (my_request) {
      Global.dbgmsg("fromRequest() + my_request");
      Global.shut_down();
    }

    if (req.stored_match_id.equals("")) {
      title = title0;
    } else {
      title = title0 + " (stored: " + req.stored_match_id + ")";
    }

    orig_req = req.createCopy();

    game_id = new String(req.game_id);
    
    player = req.player + " " + req.mode + " " + req.rating;

    if (gg.type_from_string(req.game_type) != null) {
      Global.dbgmsg("illegal game type!");
    }

    board_type = gg.get_board_type();

    if (gg.board_type_index(board_type) < 0) return false; // board type not handled

    synchro = gg.has_synchro() && gg.is_synchro_game();
    komi = gg.has_komi() && gg.is_komi_game();
    anti = gg.has_anti() && gg.is_anti_game();
    rand = gg.has_rand() && gg.is_rand_game();
    if (rand && gg.has_rand_type())
      rand_type = ""+gg.get_rand_type();
    else
      rand_type = "";
    
    black = white = false;

    if (gg.has_pref_color()) {
      if (gg.get_pref_color() == PlayerColor.BLACK) black = true;
      if (gg.get_pref_color() == PlayerColor.WHITE) white = true;
    }
    
    time  = req.my_clock.toString(true);
    otime = req.your_clock.toString(true);
    
    if (time.equals(otime)) otime = "same";

    set();
    return true;
  }

  // local state -> window
  
  private void set()
  {
    board_choice.setSelectedIndex(gg.board_type_index(board_type));

    setTitle(title);
    
    if (synchro_checkbox != null) synchro_checkbox.setSelected(synchro);    
    if (komi_checkbox    != null) komi_checkbox.setSelected(komi);    
    if (anti_checkbox    != null) anti_checkbox.setSelected(anti);    
    if (rand_checkbox    != null) rand_checkbox.setSelected(rand);
    if (black_checkbox   != null) black_checkbox.setSelected(black);
    if (white_checkbox   != null) white_checkbox.setSelected(white);

    if (rand_type_input != null) rand_type_input.setText(rand_type);
    player_input.setText(player);
    time_input.setText(time);
    otime_input.setText(otime);
    setOK();
  }


  // enable/disable accept/submit buttons & rand_type_input
  
  private void setOK()
  {
    Request req = getRequest();
    boolean ok = req != null;

    if (rand_type_input != null) rand_type_input.setEditable(rand);    

    // submit_button.setEnabled(ok);
    submit_ok = ok;

    if (!my_request) {

      ok = (req != null) && req.equals(orig_req);
      //accept_button.setEnabled(ok);
      accept_ok = ok;
    }

  }
 
  
  private Request getRequest()
  {
    Request req = new Request("???");
    
    // read out input fields

    StringTokenizer st = new StringTokenizer(player_input.getText());
    if (!st.hasMoreTokens()) return null;
    
    player     = st.nextToken();
    if (rand_type_input != null) rand_type  = rand_type_input.getText();
    time       = time_input.getText();
    otime      = otime_input.getText();

    String gts = "";

    // System.out.println("hs=" + gg.has_synchro() + " syn=" + synchro);

    if (gg.has_synchro() && synchro) gts += "s";
    gts += board_type;
    if (gg.has_komi() && komi) gts += "k";
    if (gg.has_rand() && rand) gts += "r";
    if (gg.has_rand_type() && rand) gts += rand_type;
    if (gg.has_anti() && anti) gts += "a";
    if (gg.has_pref_color()) {
      if (black) gts += "b";
      if (white) gts += "w";
    }

    // Global.dbgmsg("gts=" + gts);

    if (gg.type_from_string(gts) != null) return null;

    req.game_type = gg.type_to_string();
    
    st = new StringTokenizer(player);
    if (!st.hasMoreTokens()) return null;
    req.player = st.nextToken();

    //    Global.dbgmsg(":::" + time + ":::" + req.player);

    if (!req.my_clock.fromString(time)) return null;

    st = new StringTokenizer(otime);
    if (!st.hasMoreTokens()) return null;

    // Global.dbgmsg(":::" + otime + ":::");
    
    if (st.nextToken().equals("same")) {
      req.your_clock.fromString(time);
    } else {
      if (!req.your_clock.fromString(otime)) return null;
    }

    return req;
  }

  public void itemStateChanged(ItemEvent e)
  {
    Object src = e.getSource();
    
    if (src == board_choice) { 

      board_type = gg.board_type(board_choice.getSelectedIndex());
      setOK();

    } else if (src == synchro_checkbox) {

      synchro = (e.getStateChange() == ItemEvent.SELECTED);
      setOK();

    } else if (src == komi_checkbox) {

      komi = (e.getStateChange() == ItemEvent.SELECTED);
      setOK();

    } else if (src == anti_checkbox) {

      anti = (e.getStateChange() == ItemEvent.SELECTED);
      setOK();

    } else if (src == rand_checkbox) {

      rand = (e.getStateChange() == ItemEvent.SELECTED);
      if (rand_type_input != null) rand_type_input.setText("");
      setOK();

    } else if (src == black_checkbox) {

      black = (e.getStateChange() == ItemEvent.SELECTED);
      setOK();
	  
    } else if (src == white_checkbox) {

      white = (e.getStateChange() == ItemEvent.SELECTED);
      setOK();

    } else {

      Global.dbgmsg(">>> ItemEvent not handled: " + e);
    }
  } 

  public void actionPerformed(ActionEvent e) // was synchronized
  {
    Object src = e.getSource();

    if (src == rand_type_input) {

      rand_type = rand_type_input.getText();
      setOK();

    } else if (src == submit_button) {

      setOK();
      
      if (submit_ok) {

	if (my_request) {
	
	  Request req = getRequest();
	  String ask =  
	    "t " + login + " ask " +
	    req.game_type.toString() + " " +
	    req.my_clock.toString(true) + " ";
	  
	  if (!req.my_clock.equals(req.your_clock))
	    ask += req.your_clock.toString(true) + " ";
	  
	  ask += req.player;

	  Global.core.send_to_server_with_focus(login, ask);
	  setVisible(false);

	} else {

	  Request req = getRequest();
	  
	  Global.core.send_to_server(login, "t " + login + " decline " + game_id);
	  String ask =  
	    "t " + login + " ask " +
	    req.game_type.toString() + " " +
	    req.my_clock.toString(true) + " ";
	  
	  if (!req.my_clock.equals(req.your_clock))
	    ask += req.your_clock.toString(true) + " ";
	  
	  ask += req.player;
	  Global.core.send_to_server_with_focus(login, ask);
	  setVisible(false);
	}
      }

    } else if (src == accept_button) {

      setOK();
      
      if (accept_ok) {
	Global.core.send_to_server(login, "t " + login + " accept " + game_id);
	setVisible(false);
      }
	
    } else if (src == decline_button) {

      Global.core.send_to_server(login, "t " + login + " decline " + game_id);
      setVisible(false);
	
    } else if (src == open0_button) {

      Global.core.send_to_server(login, "t " + login + " decline " + game_id);
      Global.core.send_to_server(login, "t " + login + " open 0");
      setVisible(false);

    } else if (src == cancel_button) {

      setVisible(false);

    } else if (src == player_input) {

      player = player_input.getText();
      setOK();
      
    } else if (src == time_input) {

      time = time_input.getText();
      setOK();
      
    } else if (src == otime_input) {

      otime = otime_input.getText();
      setOK();

    } else {

      Global.dbgmsg(">>> ActionEvent not handled: " + e);
      
    }
  }

  public void make_visible() { make_visible(null); }
  
  public void make_visible(String s)
  {
    Geometry geom = Global.options.getGeometry(BoardGame.BGS+"rge");

    if (geom.get_x() > 0) {
      Global.dbgmsg("" + geom.get_x() + " " + geom.get_y());
      setLocation(geom.get_x(), geom.get_y());
    }

    if (s != null) player_input.setText(s);
    invalidate();
    validate();
    setVisible(true);
  }
  
}
