/* $Id: HistTextField.java,v 1.1.1.1 2006/08/21 04:26:02 orts_mburo Exp $
   
   (c) Michael Buro, licensed under the GPL2
*/

package ggs.ggsa.main;

import java.awt.*;
import java.awt.event.*;

public class HistTextField extends TextField implements KeyListener,FocusListener
{
  static final int HIST_NUM = 50;
  protected String[] history = new String[HIST_NUM];
  protected int hist_index = 0;
  protected boolean has_focus = false;
  
  public HistTextField() { init(); }
  public HistTextField(int a) { super(a); init(); }
  public HistTextField(String a) { super(a); init(); }
  public HistTextField(String a, int b) { super(a,b); init(); }      

  void init() { addFocusListener(this); addKeyListener(this); }

  public boolean has_focus() { return this.has_focus; }
  
  public void focusGained(FocusEvent e) 
  {
    has_focus = true;
  }

  public void focusLost(FocusEvent e) 
  {
    has_focus = false;
  }

  public void keyPressed(KeyEvent e) 
  {
    int code = e.getKeyCode();
    String s;

    // Global.dbgmsg("<" + KeyEvent.getKeyText(code) + ">");
    
    if (code == KeyEvent.VK_ESCAPE) setText("");

    else if (code == KeyEvent.VK_UP) {

      if (hist_index == 0) history[0] = getText();

      if (hist_index < history.length-1 && history[hist_index+1] != null) {
	hist_index++;
	s = history[hist_index];
	if (s == null) s = "";
	setText(s);
	setCaretPosition(Integer.MAX_VALUE);
      } else
	Toolkit.getDefaultToolkit().beep();
      
    } else if (code == KeyEvent.VK_DOWN) {      

      if (hist_index > 0) {
	hist_index--;
	s = history[hist_index];
	if (s == null) s = "";
	setText(s);
	setCaretPosition(Integer.MAX_VALUE);
      } else
	Toolkit.getDefaultToolkit().beep();

    } else if (code == KeyEvent.VK_K && e.isControlDown()) {

      // delete all characters to the right

      setText(getText().substring(0, getCaretPosition()));
      setCaretPosition(Integer.MAX_VALUE);

    } else if (code == KeyEvent.VK_D && e.isControlDown()) {

      int pos = getCaretPosition();
      String t = getText();
      
      if (pos < t.length()) {
	setText(t.substring(0, pos) + t.substring(pos+1, t.length()));
	setCaretPosition(pos);
      }

    } else if (code == KeyEvent.VK_A && e.isControlDown()) {

      setCaretPosition(0);

    } else if (code == KeyEvent.VK_E && e.isControlDown()) {

      setCaretPosition(Integer.MAX_VALUE);
    }    
  }

  public void keyReleased(KeyEvent e) {}
  public void keyTyped(KeyEvent e) {}  

  public void push()
  {
    // Global.dbgmsg("PUSH: " + getText());
    
    // append string to history
    
    if (!getText().equals(history[1])) {
      for (int i=history.length-2; i >= 0; i--)
	history[i+1] = history[i];
      history[1] = getText();
    }
    
    hist_index = 0;
    setText("");
  }
 
}
