/* $Id: PieceImages.java,v 1.3 2006/08/27 06:28:58 orts_mburo Exp $
   
   (c) Michael Buro, licensed under the GPL2
*/

package ggs.ggsa.main;

import java.awt.*;
import java.net.*;
import java.util.*;

// handles differently sized images of a playing piece

public class PieceImages
{
  private int[] sizes;
  private Image[] images;
  private boolean[] loaded;
  private String path_template;  // e.g. gif/chess/%/pdd%.gif, % gets replaced by size

  public PieceImages(String path_template, String size_string) {

    this.path_template = path_template;
    this.sizes = sizes;

    int n = 0;
    boolean f = true;
    
    // count sizes
    StringTokenizer st = new StringTokenizer(size_string, ",", false);

    while (st.hasMoreTokens() && f) {
      st.nextToken();
      ++n;
    }

    if (n == 0) { Global.dbgmsg("no size found"); Global.shut_down(); }
    
    sizes = new int[n];

    n = 0;
    st = new StringTokenizer(size_string, ",", false);

    while (st.hasMoreTokens()) {
      String s = st.nextToken();
      int sz = 0;
      try { sz = Integer.parseInt(s); }
      catch (NumberFormatException e) {
	Global.dbgmsg("PieceImages: illegal size");
	Global.shut_down();
      }

      if (sz < 1) {
	Global.dbgmsg("PieceImages: size < 1");
	Global.shut_down();
      }
	
      if (n > 0 && sz <= sizes[n-1]) {
	Global.dbgmsg("PieceImages: sizes not increasing");
	Global.shut_down();
      }

      sizes[n++] = sz;
    }    

    loaded = new boolean[sizes.length];
    images = new Image[sizes.length];    
  }
  

  private static Image read_image(String name, MediaTracker mt)
  {
    Global.dbgmsg("Reading image " + name);
    Image img = null;
    URL url = GGSEventThread.class.getResource(name);
    if (url == null)
      return null;
      
    img = Toolkit.getDefaultToolkit().getImage(url);

    mt.addImage(img, 1);

    try {
      mt.waitForID(1); //use image's ID
    } catch (InterruptedException e) {
      //Some kind of error occurred:
      Global.dbgmsg(">>> gif read error!");
      Global.shut_down();
      return null;
    }

    return img;
  }

  public synchronized int get_image_index(int size) {

    int closest = 0;
    int min_diff = Math.abs(size - sizes[0]);
    
    for (int i=1; i < sizes.length; ++i) {
      int d = Math.abs(size - sizes[i]);
      if (d < min_diff) { closest = i; min_diff = d; }
    }

    return closest;
  }

  public synchronized int get_pref_size(int size) {

    int closest = get_image_index(size);
    int min_diff = Math.abs(size - sizes[closest]);

    if (min_diff == 1 || min_diff == 2) return sizes[closest];
    else return size;
  }

  public static String expand(String template, char c, String replacement)
  {
    String s = template;

    if (replacement.indexOf(c) >= 0) return null;

    while (true) {
      int i = s.indexOf(c);
      if (i < 0) break;
      s = s.substring(0, i) + replacement + s.substring(i+1);
    }

    return s.toString();
  }
  
  
  public synchronized Image get(int index, MediaTracker mt) {

    if (!loaded[index]) {

      // load on demand
      
      loaded[index] = true;

      String name = expand(path_template, '#', ""+sizes[index]);
      
      // replace '%' by size

      images[index] = read_image(name, mt);

      if (images[index] == null) {
        Global.dbgmsg(name + " not found");
        System.exit(20);
      }
    }
    
    return images[index];
  }

}

