/* $Id: VertiLayout.java,v 1.1.1.1 2006/08/21 04:26:01 orts_mburo Exp $
   
   (c) Michael Buro, licensed under the GPL2
*/

package ggs.ggsa.main;

import java.lang.Error;
import java.awt.*;
import java.awt.event.*;
import java.util.*;

public class VertiLayout implements LayoutManager
{
  public final static int H_PARENT = 1;  // horiz from parent
  public final static int V_PARENT = 2;  // verti from parent

  // if !H_PARENT
  public final static int H_LEFT   = 4;
  public final static int H_RIGHT  = 8;
  public final static int H_CENTER = 0;  // default

  // if V_PARENT
  public final static int V_TOP    = 16;
  public final static int V_BOTTOM = 32;
  public final static int V_SPREAD = 0;  // default

  public final static int NO_TOP_PHANTOM    = 64;
  public final static int NO_BOTTOM_PHANTOM = 128;

  public boolean is_H_PARENT()  { return (align & H_PARENT) != 0; }
  public boolean is_V_PARENT()  { return (align & V_PARENT) != 0; }

  public boolean is_H_LEFT()    { return (align & H_LEFT)  != 0;    }
  public boolean is_H_RIGHT()   { return (align & H_RIGHT) != 0; }
  public boolean is_H_CENTER()  { return !is_H_LEFT() && !is_H_RIGHT(); }

  public boolean is_V_TOP()     { return (align & V_TOP)    != 0;   }
  public boolean is_V_BOTTOM()  { return (align & V_BOTTOM) != 0;  }
  public boolean is_V_SPREAD()  { return !is_V_TOP() && !is_V_BOTTOM(); }

  public boolean is_TOP_PHANTOM()    { return (align & NO_TOP_PHANTOM)    == 0; }
  public boolean is_BOTTOM_PHANTOM() { return (align & NO_BOTTOM_PHANTOM) == 0; }

  private final static int P_WEIGHT_DEF = 1;

  private final static int PREFERRED    = 0;
  private final static int MINIMUM      = 1;
  private final static int MAXIMUM      = 2;


  private int  align;
  private int  v_gap;
  private int  p_weight;
  private Hashtable<Component,VertiLayoutConstraints> components =
    new Hashtable<Component,VertiLayoutConstraints>();

  public VertiLayout(int align, int v_gap) { this(align, v_gap, P_WEIGHT_DEF); }

  public VertiLayout(int align, int v_gap, int p_weight)
  {
    this.align = align;
    this.v_gap = v_gap;
    this.p_weight = p_weight;
  }    

  public void addLayoutComponent(String name, Component component) 
  {
  }
	
  public void removeLayoutComponent(Component component) 
  {
    components.remove(component);
  }
	
  public void setConstraints(Component comp, VertiLayoutConstraints cons) 
  {
    if (!is_V_PARENT() && cons.is_ADJ())
      throw new Error("adj. component in !V_PARENT-container");

    try { components.put(comp, (VertiLayoutConstraints) cons.clone() ); }
    catch( CloneNotSupportedException e ) { }
  }

  public VertiLayoutConstraints getConstraints(Component comp) 
  {
    VertiLayoutConstraints cons = components.get(comp);

    if (cons != null) {

      try { return (VertiLayoutConstraints) cons.clone(); }
      catch(CloneNotSupportedException e ) { return null; }
                
    } else return null;
  }


  public Dimension minimumLayoutSize(Container parent) 
  { 
    return calcLayoutSize(parent, MINIMUM); 
  }

  public Dimension preferredLayoutSize(Container parent) 
  { 
    return calcLayoutSize(parent, PREFERRED);
  }

  private Dimension calcLayoutSize(Container parent, int size_type) 
  {
    int max_x = 0, sum_y = 0;
    Dimension comp_size = null;
    Dimension cont_size = parent.getSize();

    //System.out.println("calcsize " + parent.getSize().width + " "
    //		   + parent.getSize().height + " " + size_type + " ");

    for (int counter = 0; counter < parent.getComponentCount(); counter++) {

      Component comp = parent.getComponent(counter);
      Dimension dim = comp.getSize();

      /* 
      if (is_H_PARENT()) {
	dim.width = cont_size.width - parent.getInsets().right - parent.getInsets().left;
      }

      if (is_V_PARENT()) {
	dim.height = cont_size.height - parent.getInsets().top - parent.getInsets().bottom;
      }

      comp.setSize(dim);

      */

      if (size_type == PREFERRED) { 

	comp_size = comp.getPreferredSize(); 
	
      } else if (size_type == MINIMUM) { 
	
	comp_size = comp.getMinimumSize(); 
	
      } else { 
	
	comp_size = comp.getMaximumSize(); 
      }
      
      if (comp_size.width > max_x) max_x = comp_size.width;
      
      sum_y += comp_size.height + v_gap;
    }
    
    sum_y -= v_gap;
    
    max_x += parent.getInsets().left + parent.getInsets().right;
    sum_y += parent.getInsets().top  + parent.getInsets().bottom;

    if (is_H_PARENT()) max_x = parent.getSize().width;
    if (is_V_PARENT()) sum_y = 0; // parent.getSize().height;

    //    System.out.println(" -> " + max_x + " " + sum_y);

    return new Dimension(max_x, sum_y);
  }


  public void layoutContainer(Container parent) 
  {
    //    System.out.println("verti layout " + 
    //		       parent.getSize().width + " " + parent.getSize().height);

    // Dimension dim = calcLayoutSize(parent, PREFERRED);

    int cw = parent.getSize().width  - parent.getInsets().left - parent.getInsets().right;
    int ch = parent.getSize().height - parent.getInsets().top  - parent.getInsets().bottom;

    if (is_V_TOP() || is_V_BOTTOM()) {

      //      System.out.println("V_TOP | V_BOTTOM");

      int cx, cy;

      if (is_V_BOTTOM()) {
	int total_height = 0;

	for(int counter = 0; counter < parent.getComponentCount(); counter++) {
	  total_height += parent.getComponent(counter).getPreferredSize().height + v_gap;
	}
	
	total_height -= v_gap;
	cy = cw-total_height+parent.getInsets().top;

      } else {

	cy = parent.getInsets().top;

      }

      cx = parent.getInsets().left;

      for(int counter = 0; counter < parent.getComponentCount(); counter++) {
	
	Component comp      = parent.getComponent(counter);
	Dimension comp_size = comp.getPreferredSize();
	Point     comp_loc  = comp.getLocation();
	
	int x, width, height;
	
	if (is_H_PARENT()) {
	  
	  width = cw;
	  x = cx;
	  
	} else {
	  
	  width = comp_size.width;
	  
	  if      (is_H_CENTER()) x = cx + (cw - width)/2;
	  else if (is_H_LEFT())   x = cx;
	  else                    x = cx + width - (cw+1)/2;

	}

	//	System.out.println("comp " + (counter+1) + " " + 
	//			   x + " " + cy + " " + width + " " + comp_size.height);
	//	System.out.println("cw= " + cw + " " + " width=" + width);

	comp.setLocation(x, cy);
	comp.setSize(width, comp_size.height);
	// comp.doLayout();
	
	cy += comp_size.height + v_gap;
      }

    } else {

      // V_SPREAD

      //      System.out.println("V_SPREAD");

      int fixed_height = 0;
      int w_sum = 0;

      for (int counter = 0; counter < parent.getComponentCount(); counter++) {
	Component comp = parent.getComponent(counter);
	VertiLayoutConstraints cons = getConstraints(comp);

	if (cons == null) throw new Error("no veri_constraint found");

	if (cons.is_FIX()) fixed_height += comp.getPreferredSize().height;
	else w_sum += cons.get_ADJ_weight();

	w_sum += cons.get_glue_weight();
	fixed_height += v_gap;
      }
      
      fixed_height -= v_gap;

      //System.out.println(">> " + is_TOP_PHANTOM() + " " + is_BOTTOM_PHANTOM());

      if (is_TOP_PHANTOM())    w_sum += p_weight;
      if (is_BOTTOM_PHANTOM()) w_sum += p_weight;

      int adj_height = ch - fixed_height;
      
      //      System.out.println("adj=" + adj_height + " fix=" + fixed_height + " ch=" + ch + " w=" + w_sum);
     

      int cx = parent.getInsets().left;
      int cy = parent.getInsets().top;

      if (is_TOP_PHANTOM()) 
	cy += Math.round((p_weight/(double)w_sum)*adj_height);

      for(int counter = 0; counter < parent.getComponentCount(); counter++) {
	
	Component comp = parent.getComponent(counter);
	Dimension comp_size = comp.getPreferredSize();
	Point     comp_loc  = comp.getLocation();
	
	int x, width, height;
	
	if (is_H_PARENT()) {
	  
	  width = cw;
	  x = cx;
	  
	} else {
	  
	  width = comp_size.width;
	  
	  if      (is_H_CENTER()) x = cx + (cw - width)/2;
	  else if (is_H_LEFT())   x = cx;
	  else                    x = cx + width - (cw+1)/2;
	  
	}

	VertiLayoutConstraints cons = getConstraints(comp);
	if (cons == null) throw new Error("no veri_constraint found");

	if (cons.is_ADJ()) {
	  height = (int)Math.round((cons.get_ADJ_weight()/(double)w_sum)*adj_height);
	  //	  System.out.println(" / " + adj_height);

	} else {
	  height = comp_size.height;
	}

	comp.setLocation(x, cy);
	comp.setSize(width, height);

	//	System.out.println("comp " + (counter+1) + " " + 
	//			   x + " " + cy + " " + width + " " + height);

	cy += height + v_gap;

	int glue = cons.get_glue_weight();

	//System.out.println("glue=" + glue);

	if (glue > 0) cy += Math.round((glue/(double)w_sum)*adj_height);
      }
    }

    //System.out.println("done.\n");


  }
}
