/* $Id: SocketComm.java,v 1.3 2006/08/29 09:11:24 orts_mburo Exp $
   
   (c) Michael Buro, licensed under the GPL2
*/

package ggs.service.common;

import java.awt.*;
import java.io.*;
import java.net.*;
import java.util.*;
import ggs.shared.*;

class SocketComm
{
  final boolean READ_LINE   = false;
  final boolean USE_TIMEOUT = true;  // true => READ_LINE=false
  final int     TIMEOUT     = 4000;  // millis
  
  private String           host   = null;
  private int              port   = 0;
  private Socket           socket = null;
  private InputStream      input = null;
  private OutputStream     output = null;
  
  private String last_sent_line = "";
  
  public void connect(String host_, int port_,
		      String name, String password)
    throws IOException, UnknownHostException
  {
    Misc.msg("trying to connect " + name + "@" + host_ + ":" + port_);

    host = host_;
    port = port_;
    socket = new Socket(host, port);

    if (USE_TIMEOUT) {
    
      // set a socket timeout

      try { socket.setSoTimeout(TIMEOUT); }
      catch (SocketException se) {
	Misc.err("Unable to set socket option SO_TIMEOUT");
      }
    }
    
    // output = new DataOutputStream(socket.getOutputStream());
    // input = new BufferedReader(new InputStreamReader(socket.getInputStream()));

    output = socket.getOutputStream();
    input = socket.getInputStream();
    
    Misc.msg("connected");

    String line;
    boolean login_sent = false;
    
    for (;;) {
      line = read_line();
      if (line.endsWith("use).")) {

	if (login_sent) {
	  Misc.msg("login sent twice - giving up");
	  throw new IOException("illegal login name");
	}
	  
	if (!name.equals("")) write_line(name);
	login_sent = true;
	
      } else if (line.indexOf("password") >= 0) {

	Misc.msg("LOGIN= " + last_sent_line);
	Global.options.put2("lo", last_sent_line); // set login
	
	if (!password.equals("")) write_line(password);
	continue;
	
      } else if (line.indexOf("READY") >= 0) break;

    }
  }


  public void write_line(String line) throws IOException
  {
    synchronized (output) {

      last_sent_line = line; // needed during login
      
      if (output != null) {

	byte[] bytes = line.getBytes("UTF-8"); 

	while (true) {
	  try { output.write(bytes); }
	  catch (InterruptedIOException iioe) {
	    Misc.msg("write TIMEOUT");
	    continue;
	  }
	  break;
	}

	write_byte((byte)'\n');
	output.flush();
	
      } else Misc.err("socket problem");
    }
  }


  private void write_byte(byte b) throws IOException
  {
    int ci;

    while (true) {
      try { output.write(b); }
      catch (InterruptedIOException iioe) {
	Misc.msg("write TIMEOUT");
	continue;
      }
      break;
    }
  }

  private int read_byte() throws IOException
  {
    int ci;

    while (true) {
      try { ci = input.read(); }
      catch (InterruptedIOException iioe) {
	//Misc.msg("read TIMEOUT");
	continue;
      }
      break;
    }

    return ci;
  }

  
  public String read_line() throws IOException
  {
    synchronized (input) {
    
      if (input != null) {

	int ci;
	Vector<Byte> vb = new Vector<Byte>();
	  
	do {

	  ci = read_byte();
	  
	  if (ci == 13) {  // CR
	    ci = read_byte();
	    if (ci == 10)
	      break; // LF
	  }
	  
	  if (ci < 0) {
	    Misc.err("socket EOF");
	    return "";
	  }
	  
	  vb.add(new Byte((byte)ci));
	  
	} while (true);

	byte[] bytes = new byte[vb.size()];

	for (int i=0; i < vb.size(); ++i)
	  bytes[i] = vb.get(i);

	String s = new String(bytes, "UTF-8");
	return s;
      }

      Misc.err("socket problem");

      return null;
    }
  }
  
  public void close()
  {
    try {

      if (socket != null) {
	socket.close(); socket = null;
      }

      if (input != null) {
	input.close();  input  = null;
      }

      if (output != null) {
	output.close(); output = null;
      }

    }
    catch (IOException e) { }
  }

}

