/** 
    $Id: SkatGame.java,v 1.9 2006/09/04 03:40:42 orts_mburo Exp $
    $Source: /usr/bodo1/cvs/cvssoftware/ggs/src/ggs/service/skat/SkatGame.java,v $

    Skat Game Implementation, wraps SkatState

    (c) Michael Buro, licensed under GPL2
*/

package ggs.service.skat;

import ggs.service.common.*;
import ggs.shared.*;
import java.util.*;


public class SkatGame extends SkatState implements Game, TableCreator
{
  /**
     @return game name
  */
  public String getName() { return "skat"; }

  /**
     @return game login
  */
  public String getLogin() { return "/skat"; }

  /**
     @return new game
  */
  public Game newGame() { return new SkatGame(); }
  
  /**
     @return new table
  */
  public Table newTable() 
  {
    return new StdTable(new SkatMatch(newGame()));
  }

  // ----- test code ---------------------------------
  
  /** run some tests */
  public void test()
  {
    SkatGame sg = new SkatGame();
    Random gen = new Random();

    Vector<String> msgs = new Vector<String>();
    sg.initState("regular", 3, gen);

    for (;;) {
      if (tm(sg, "18"))   break;
      if (tm(sg, "YES"))  break;
      if (tm(sg, "20"))   break;
      if (tm(sg, "YES"))  break;
      if (tm(sg, "22"))   break;
      if (tm(sg, "NO"))   break;
      if (tm(sg, "24"))   break;
      if (tm(sg, "YES"))  break;
      if (tm(sg, "PASS")) break;
      if (tm(sg, "SKAT")) break;

      while (!sg.finished()) {
	String rm = sg.random_move(gen);
	if (tm(sg, rm))
	  continue;
      }

      if (sg.finished())
	Misc.msg("OK");
      else
	Misc.msg("PROBLEM");	
      break;
    }
  }

  /** generate random move (only discard, game declaration, and card play) */
  public String random_move(Random gen)
  {
    if (gen.nextInt(10) < 1)
      return "foo";

    if (phase == SkatState.Phase.CARDPLAY) {

      Vector<Card> cards = new Vector<Card>();

      for (Card c : hands[to_move].cards)
	cards.add(c);
      
      Card c = cards.get(gen.nextInt(cards.size()));
      return c.toString();
    }

    if (phase == SkatState.Phase.DISCARD) {

      Vector<Card> cards = new Vector<Card>();

      cards.addAll(hands[to_move].cards);
      cards.add(skat.get(0));
      cards.add(skat.get(1));

      for (Card c : cards) {
	System.out.print(c.toString() + " ");
      }

      System.out.println("to_move:" + to_move);
      
      // choose two random cards from hand + skat
      int r1 = gen.nextInt(cards.size());
      int r2;

      do {
	r2 = gen.nextInt(cards.size());
      } while (r1 == r2);

      return cards.get(r1).toString() + "," + cards.get(r2).toString() + ",G";
    }

    return "";
  }

  /** test move */
  static public boolean tm(SkatGame g, String move)
  {
    Vector<String> msgs = new Vector<String>();

    Misc.msg("\nmove: " + g.to_move + " " + move);
    String res = g.makeMove(g.to_move, move, true, false);

    if (res != null) {
      Misc.msg("error: " + res);
      return true;
    }

    Misc.msg("messages:");
    for (String s : msgs) {
      Misc.msg(s);
    }

    return false;
  }
  
}
